# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Jonas Mertens

from __future__ import annotations

import json
from collections.abc import Iterable, Mapping
from typing import Any

from PyQt5.QtCore import QItemSelectionModel, Qt, QTimer, pyqtSignal, pyqtSlot
from PyQt5.QtGui import (
    QBrush,
    QColor,
    QDropEvent,
    QFocusEvent,
    QKeyEvent,
    QMouseEvent,
    QStandardItem,
    QStandardItemModel,
)
from PyQt5.QtWidgets import (
    QAction,
    QApplication,
    QDesktopWidget,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QSplitter,
    QTreeView,
    QVBoxLayout,
    QWidget,
)
from qcodes.instrument.channel import InstrumentModule
from qcodes.instrument.instrument import Instrument
from qcodes.instrument.parameter import Parameter
from qcodes.utils.metadata import Metadatable

from qumada.instrument.mapping.base import TerminalParameters, filter_flatten_parameters
from qumada.metadata import Metadata

RED = QColor(255, 0, 0)
WHITE = QColor(255, 255, 255)
GREEN = QColor(0, 255, 0)
YELLOW = QColor(255, 255, 0)
PINK = QColor(255, 192, 203)
BLUE = QColor(0, 0, 255)


# TODO: terminal_parameter attributes
class TerminalTreeView(QTreeView):
    """
    QTreeView, that displays QuMADA `TerminalParameters` datastructure
    (`Mapping[Any, Mapping[Any, Parameter] | Parameter]`).
    Items are draggable to map them to instruments.
    """

    def __init__(self, monitoring=False):
        super().__init__()

        model = QStandardItemModel()
        model.setHorizontalHeaderLabels(["Terminal", "Mapped Instrument", "Monitoring"])
        self.setModel(model)
        self.terminal_parameters = None

        self.selected_terminal_tree_elem = None

        # Parameter Monitoring Functionality
        self.monitoring_enable = monitoring
        self.monitoring_timer = QTimer()
        self.monitoring_timer.timeout.connect(self.update_monitoring)

        self.setDragEnabled(True)

    # forward keypressevents "hack". keyPressEvent of QMainWindow doesnt fire for letter keys
    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() == Qt.Key_A or event.key() == Qt.Key_R or event.key() == Qt.Key_E or event.key() == Qt.Key_U:
            self.parentWidget().keyPressEvent(event)
        else:
            return super().keyPressEvent(event)

    # overwrite default double click event behaviour
    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        pass

    def update_monitoring(self):
        """
        This is called periodically, gets the mapped parameter values and updates the labels
        """
        # TODO: maybe format cells based on settable
        # if(param.settable):
        #     pass # maybe do coloring based on settable

        root = self.model().invisibleRootItem()
        for terminal in get_children(root):
            for terminal_param in get_children(terminal):
                param = self.terminal_parameters[terminal_param.source[0]][terminal_param.source[1]]
                if param is not None:
                    if self.monitoring_get_type == "get" and param.gettable:
                        # Use get command
                        try:
                            val = param.get()
                        except Exception:
                            val = param.cache.get(get_if_invalid=False)
                    else:
                        # Use cached value (also applicable to non-gettable parameters (last set value))
                        val = param.cache.get(get_if_invalid=False)

                    if val is not None and (type(val) is int or type(val) is float):
                        if param.unit is not None:
                            self.model().setData(terminal_param.index().siblingAtColumn(2), f"{val:.2f} {param.unit}")
                        else:
                            self.model().setData(terminal_param.index().siblingAtColumn(2), f"{val:.2f}")
                else:
                    self.model().setData(terminal_param.index().siblingAtColumn(2), "")

        self.resizeColumnToContents(2)

    def update_tree(self):
        """
        Function that sets all visual elements (labels, colors) based on state of mapping.
        This is a clean solution (not super efficient) because changing the mapping for one
        parameter can have an influence on any other visual elements (duplicate markings etc.)
        """
        # Set mapped instrument labels and colors, set colors for duplicates
        tree = self.model()
        parameter_duplicates = {}
        for terminal in get_children(tree.invisibleRootItem()):
            tree.setData(terminal.index(), QBrush(WHITE), Qt.BackgroundRole)
            tree.setData(terminal.index().siblingAtColumn(1), QBrush(RED), Qt.BackgroundRole)
            tree.setData(terminal.index().siblingAtColumn(1), "")
            all_mapped = True
            any_mapped = False
            channel_names = set()
            for terminal_param in get_children(terminal):
                tree.setData(terminal_param.index(), QBrush(WHITE), Qt.BackgroundRole)
                tree.setData(terminal_param.index().siblingAtColumn(1), QBrush(RED), Qt.BackgroundRole)
                tree.setData(terminal_param.index().siblingAtColumn(1), "")
                param = self.terminal_parameters[terminal_param.source[0]][terminal_param.source[1]]
                if param is not None:
                    if isinstance(param, Parameter):
                        param_hash = hash(param)
                        if param_hash in parameter_duplicates:
                            parameter_duplicates[param_hash].append(terminal_param)
                        else:
                            parameter_duplicates[param_hash] = [terminal_param]

                        channel_names.add(param.instrument.full_name)
                        tree.setData(terminal_param.index().siblingAtColumn(1), param.full_name)
                        tree.setData(terminal_param.index().siblingAtColumn(1), QBrush(GREEN), Qt.BackgroundRole)
                        any_mapped = True
                    else:
                        raise TypeError("Gate parameters have to be either None or of type Parameter.")
                else:
                    all_mapped = False

            # color and label terminals
            if all_mapped:
                tree.setData(terminal.index().siblingAtColumn(1), QBrush(GREEN), Qt.BackgroundRole)
                if len(channel_names) == 1:
                    tree.setData(terminal.index().siblingAtColumn(1), channel_names.pop())
                else:
                    tree.setData(terminal.index().siblingAtColumn(1), "")
            elif any_mapped:
                tree.setData(terminal.index().siblingAtColumn(1), QBrush(YELLOW), Qt.BackgroundRole)

        # color duplicates
        for items_with_param in parameter_duplicates.values():
            if len(items_with_param) != 1:
                for duplicate in items_with_param:
                    duplicate.setData(QBrush(PINK), Qt.BackgroundRole)
                    duplicate.parent().setData(QBrush(PINK), Qt.BackgroundRole)
                    # tree.setData(duplicate.index(), QBrush(PINK), Qt.BackgroundRole)
                    # tree.setData(duplicate.parent().index(), QBrush(PINK), Qt.BackgroundRole)

        # any item selected?
        if self.selected_terminal_tree_elem is not None:
            self.selected_terminal_tree_elem.setData(QBrush(BLUE), Qt.BackgroundRole)
            # tree.setData(self.selected_terminal_tree_elem, QBrush(BLUE), Qt.BackgroundRole)

        self.resizeColumnToContents(1)
        self.update_monitoring()

    def import_data(self, terminal_parameters: TerminalParameters) -> None:
        """Build up tree with provided terminal parameters."""
        root = self.model().invisibleRootItem()
        self.terminal_parameters = terminal_parameters
        for terminal_name, terminal_params in terminal_parameters.items():
            item = QStandardItem(terminal_name)
            # item.setData((terminal_name, tuple(terminal_params.keys())))
            # print(item.data(),item.text())
            item.source = (terminal_name, tuple(terminal_params.keys()))
            root.appendRow(item)

            # if isinstance(terminal_params, Mapping):
            for terminal_param_name in terminal_params.keys():
                subitem = QStandardItem(terminal_param_name)
                # subitem.setData((terminal_name, terminal_param_name))
                # print(subitem.data(),subitem.text())
                subitem.source = (terminal_name, terminal_param_name)
                item.appendRow(subitem)

            qidx = item.index()
            self.model().setData(qidx.siblingAtColumn(1), QBrush(RED), Qt.BackgroundRole)
            self.model().insertColumn(1, qidx)
            self.model().insertColumn(2, qidx)
            for i in range(len(terminal_params.keys())):
                self.model().setData(qidx.child(i, 1), "")
                self.model().setData(qidx.child(i, 1), QBrush(RED), Qt.BackgroundRole)
                self.model().setData(qidx.child(i, 2), "")

            self.setColumnHidden(2, not self.monitoring_enable)


class InstrumentTreeView(QTreeView):
    """QTreeView, that displays qcodes instruments."""

    drag_terminal_drop_instr = pyqtSignal(QStandardItem, QStandardItem)

    def __init__(self):
        super().__init__()

        model = QStandardItemModel()
        # model.setHorizontalHeaderLabels(["Name", "Terminal"])
        model.setHorizontalHeaderLabels(["Instrument Name"])
        self.setModel(model)

        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        pass

    def focusOutEvent(self, a0: QFocusEvent) -> None:
        mainwindow = self.parentWidget()
        for child in mainwindow.children():
            if isinstance(child, TerminalTreeView):
                terminal_tree = child
                break

        if terminal_tree.selected_terminal_tree_elem is not None:
            terminal_tree.selected_terminal_tree_elem = None
            terminal_tree.update_tree()

        return super().focusOutEvent(a0)

    # forward keypressevents "hack". keyPressEvent of QMainWindow doesnt fire for letter keys
    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() == Qt.Key_A or event.key() == Qt.Key_R or event.key() == Qt.Key_E or event.key() == Qt.Key_U:
            self.parentWidget().keyPressEvent(event)
        else:
            return super().keyPressEvent(event)

    def get_perfect_mappings(self, terminal_params: list[str], parent_elem=None) -> list[QStandardItem]:
        """
        Given a list of terminal_parameters (names) return a list of QStandardItems in InstrumentTree
        (perfect mapping candidate) that could be mapped perfectly i.e. every terminal_parameter can be
        mapped uniquely to the list of all children parameters of the perfect mapping candidate.
        """
        if parent_elem is None:
            parent_elem = self.model().invisibleRootItem()

        perfect_elems = []
        for instrument_elem in get_children(parent_elem):
            if instrument_elem.hasChildren():
                all_mappings = self.get_all_possible_mapping_names(instrument_elem)
                perfect_mapping = True
                for terminal_param in terminal_params:
                    if all_mappings.count(terminal_param) != 1:
                        perfect_mapping = False
                        break

                if perfect_mapping:
                    perfect_elems.append(instrument_elem)
                if instrument_elem.hasChildren():
                    perfect_elems = perfect_elems + self.get_perfect_mappings(terminal_params, instrument_elem)

        return perfect_elems

    def get_all_possible_mapping_names(self: InstrumentTreeView, instrument_elem: QStandardItem) -> list[str] | None:
        """
        Get all possible mapping names for all children of instrument_elem. None if
        """
        mapping_names = []
        if not hasattr(instrument_elem, "source"):
            return None

        if instrument_elem.hasChildren():
            for child in get_children(instrument_elem):
                mapping_names = mapping_names + self.get_all_possible_mapping_names(child)
        else:
            if isinstance(instrument_elem.source, Parameter):
                if hasattr(instrument_elem.source, "_mapping"):
                    return [instrument_elem.source._mapping]

        return mapping_names

    def map_given_terminal_instrument_elem_selection(
        self,
        tree: TerminalTreeView,
        terminal_tree_traversal: tuple[str, tuple[str]] | tuple[str, str],
        instr_elem: Metadatable | Parameter,
    ) -> bool:
        """
        For a selected item in terminal_tree (given via terminal_tree_traversal) and selected item in instrument_tree
        (instr_elem) do the mapping process. Behaviour based on combinations like: direct mapping between parameters,
        automap to all children etc.
        """
        mapped = False
        if isinstance(terminal_tree_traversal[1], str) and isinstance(instr_elem, Parameter):
            # map directly - should mapping be forbidden if _mapping attribute of Parameter does not fit?
            tree.map_parameter(instr_elem, terminal_tree_traversal)
            mapped = True
            # maybe later - if used this should be done somewhere else (map_parameter)
            # self.add_terminal_to_view(parent, row, f"{terminal[0]}.{terminal[1]}")
        elif isinstance(terminal_tree_traversal[1], tuple) and isinstance(instr_elem, (InstrumentModule, Instrument)):
            # map automatically as much as possible
            all_params = filter_flatten_parameters(instr_elem)
            child_params_all = {
                param_name: param
                for param_name, param in all_params.items()
                if instr_elem in param.instrument.ancestors
            }

            # try to map chosen terminal_parameters with child parameters of instr_elem if uniquely possible
            possible_mappings = get_possible_mapping_candidates(terminal_tree_traversal[1], child_params_all)
            for terminal_param_name, parameter_candidates in possible_mappings.items():
                if len(parameter_candidates) == 1:
                    # unique mapping - iterate over terminal_parameters of terminal

                    terminal = get_child(tree.model().invisibleRootItem(), terminal_tree_traversal[0])
                    terminal_param = get_child(terminal, terminal_param_name)
                    tree.map_parameter(parameter_candidates[0], terminal_param.source)
                    mapped = True

        elif isinstance(terminal_tree_traversal[1], str) and isinstance(instr_elem, (InstrumentModule, Instrument)):
            # map automatically to one parameter
            all_params = filter_flatten_parameters(instr_elem)
            child_params_all = {
                param_name: param
                for param_name, param in all_params.items()
                if instr_elem in param.instrument.ancestors
            }
            possible_mappings = get_possible_mapping_candidates((terminal_tree_traversal[1],), child_params_all)

            # dict is always of size 1 (only one terminal_parameter)
            terminal_param_name = terminal_tree_traversal[1]
            possible_instrument_parameters = possible_mappings[terminal_param_name]
            if len(possible_instrument_parameters) == 1:
                if terminal_param_name == terminal_tree_traversal[1]:
                    tree.map_parameter(possible_instrument_parameters[0], terminal_tree_traversal)
                    mapped = True
        elif isinstance(terminal_tree_traversal[1], tuple) and isinstance(instr_elem, Parameter):
            # map parameter of chosen terminal to chosen instrument parameter
            parameter = instr_elem
            terminal_parameters = terminal_tree_traversal[1]
            if hasattr(parameter, "_mapping"):
                if parameter._mapping in terminal_parameters:
                    tree.map_parameter(parameter, (terminal_tree_traversal[0], parameter._mapping))
                    mapped = True

        return mapped

    def dropEvent(self, event: QDropEvent) -> None:
        """
        Start mapping based on selected element in terminal tree (dragged from)
        and selected element in instrument tree (dropped to).

        Different mapping behaviour depending on the combination of "types" of elements
        (see map_given_terminal_instrument_elem_selection).
        """
        dest_index = self.indexAt(event.pos())
        if not dest_index.isValid():
            return

        if dest_index.column() != 0:
            instr_elem = self.model().itemFromIndex(dest_index.siblingAtColumn(0))
        else:
            instr_elem = self.model().itemFromIndex(dest_index)

        tree = event.source()
        assert isinstance(tree, TerminalTreeView)
        terminal_tree_idx = tree.currentIndex().siblingAtColumn(0)
        terminal_elem = tree.model().itemFromIndex(terminal_tree_idx)

        # communicate selected instrument/terminal to main window.
        # drag_terminal_drop_instr_slot catches this signal and does mapping
        self.drag_terminal_drop_instr.emit(instr_elem, terminal_elem)

    def add_terminal_to_view(self, parent, row, terminal_name):
        model = self.model()
        # Create cell, if its not there yet
        if not model.hasIndex(row, 1, parent):
            model.insertColumn(1, parent)
        # Add terminal name to instrument row
        model.setData(parent.child(row, 1), terminal_name)

    def import_data(self, components: Mapping[Any, Metadatable]) -> None:
        """Build up tree with provided instruments."""
        parent = self.model().invisibleRootItem()
        seen: set[int] = set()

        def recurse(node, parent) -> None:
            """Recursive part of the function. Fills instrument_parameters dict."""
            # TODO: Change this try-except-phrase to match-case, when switched to Python3.10
            try:
                values = list(node.values()) if isinstance(node, dict) else list(node)
            except KeyError:
                values = [node]
            except IndexError:
                values = []

            for value in values:
                if isinstance(value, Parameter):
                    item = QStandardItem(value.name)
                    item.source = value
                    parent.appendRow(item)
                else:
                    if isinstance(value, Iterable) and not isinstance(value, str):
                        recurse(value, parent)
                    elif isinstance(value, Metadatable):
                        # Object of some Metadatable type, try to get __dict__ and _filter_flatten_parameters
                        try:
                            value_hash = hash(value)
                            if parent is not self.model().invisibleRootItem():
                                try:
                                    if value in parent.source.ancestors:
                                        continue

                                    if len(value.ancestors) >= 2:
                                        if not value.ancestors[1] == parent.source:
                                            # print(f"{value.full_name} is not a direct decendant of
                                            # {parent.source.full_name}")
                                            continue
                                except AttributeError:
                                    continue

                            if value_hash not in seen:
                                seen.add(value_hash)
                                item = QStandardItem(value.name)
                                item.source = (
                                    value  # TODO: does it make sense to save value as .data()? or is .source ok?
                                )
                                parent.appendRow(item)
                                recurse(vars(value), item)
                                # if item has no parameters (item has no children after recurse) delete (last) row
                                if not item.hasChildren():
                                    num_rows = parent.rowCount()
                                    parent.removeRow(num_rows - 1)
                        except TypeError:
                            # End of tree
                            pass

        recurse(components, parent)


class MainWindow(QMainWindow):
    """
    Main window containing the two trees and buttons
    """

    def __init__(
        self,
        components,
        terminal_parameters,
        monitoring: bool = False,
    ):
        super().__init__()
        self.components = components
        self.setWindowTitle("Mapping GUI")

        container = QWidget()
        layout = QVBoxLayout()

        # Tree views
        self.instrument_tree = InstrumentTreeView()
        self.instrument_tree.import_data(components)
        self.instrument_tree.setSelectionMode(self.instrument_tree.SelectionMode(1))
        self.instrument_tree.drag_terminal_drop_instr.connect(self.drag_terminal_drop_instr_slot)
        self.instrument_tree.setSizePolicy(QSizePolicy.Policy(3), QSizePolicy.Policy(3))

        self.terminal_tree = TerminalTreeView(monitoring)
        self.terminal_tree.import_data(terminal_parameters)
        self.terminal_tree.instrument_model = self.instrument_tree.model()
        self.terminal_tree.setSelectionMode(self.terminal_tree.SelectionMode(1))
        self.terminal_tree.selected_terminal_tree_elem = None
        self.terminal_tree.expandAll()
        self.terminal_tree.resizeColumnToContents(0)
        self.terminal_tree.collapseAll()
        self.terminal_tree.resizeColumnToContents(1)
        self.terminal_tree.setSizePolicy(QSizePolicy.Policy(3), QSizePolicy.Policy(3))

        upper_widget = QSplitter(Qt.Horizontal)
        upper_widget.addWidget(self.terminal_tree)
        upper_widget.addWidget(self.instrument_tree)
        upper_widget.setChildrenCollapsible(False)

        # Buttons
        button_container = QWidget()
        button_layout = QHBoxLayout()

        self.button_map_auto = QPushButton("Map automatically (a)")
        self.button_map_auto.clicked.connect(self.map_automatically)

        self.button_reset_mapping = QPushButton("Reset mapping (r)")
        self.button_reset_mapping.clicked.connect(self.reset_mapping)

        self.button_unfold_terminals = QPushButton("Unfold (u)")
        self.button_unfold_terminals.clicked.connect(self.unfold_terminals)

        self.button_exit = QPushButton("Exit (e)")
        self.button_exit.clicked.connect(self.close)

        # Button layout
        button_layout.addWidget(self.button_map_auto)
        button_layout.addWidget(self.button_reset_mapping)
        button_layout.addWidget(self.button_unfold_terminals)
        button_layout.addStretch()
        button_layout.addWidget(self.button_exit)
        button_container.setLayout(button_layout)
        button_container.setSizePolicy(QSizePolicy.Policy(3), QSizePolicy.Policy(0))

        # Menu
        menu = self.menuBar()

        # Monitoring
        Monitoring_menu = menu.addMenu("Monitoring")
        self.monitoring_enable = monitoring
        self.terminal_tree.monitoring_get_type = "get"
        if monitoring:
            self.terminal_tree.monitoring_timer.start(1000)

        # Monitoring - Refresh delay
        self.monitoring_refresh_delay = QAction("Refresh Delay", self)
        self.monitoring_refresh_delay.triggered.connect(self.set_refresh_rate)
        Monitoring_menu.addAction(self.monitoring_refresh_delay)

        # Monitoring - Enable/Disable
        if monitoring:
            self.toggle_monitoring_action = QAction("Disable", self)
        else:
            self.toggle_monitoring_action = QAction("Enable", self)

        self.toggle_monitoring_action.triggered.connect(self.toggle_monitoring)
        Monitoring_menu.addAction(self.toggle_monitoring_action)

        # Monitoring - Get type (from cache or by get() command)
        get_type_menu = Monitoring_menu.addMenu("Get type")
        self.use_cache_action = QAction("Only Cached values", self)
        self.use_get_action = QAction("Get command", self)
        self.use_cache_action.setCheckable(True)
        self.use_get_action.setCheckable(True)
        self.use_get_action.setChecked(True)
        get_type_menu.addAction(self.use_cache_action)
        get_type_menu.addAction(self.use_get_action)
        self.use_get_action.triggered.connect(self.monitoring_set_get_type)
        self.use_cache_action.triggered.connect(self.monitoring_set_cache_type)

        # Help button
        help_action = menu.addAction("Help")
        help_action.triggered.connect(self.show_help)

        # Main layout
        layout.addWidget(upper_widget)
        layout.addWidget(button_container)

        container.setLayout(layout)
        self.setCentralWidget(container)
        self.terminal_tree.setFocus()
        idx = self.terminal_tree.model().invisibleRootItem().child(0, 0).index()
        self.terminal_tree.selectionModel().select(idx, QItemSelectionModel.SelectionFlag.ClearAndSelect)
        self.terminal_tree.setCurrentIndex(idx)
        self.resize(QDesktopWidget().availableGeometry(self).size() * 0.45)

        self.terminal_parameters = terminal_parameters

        self.terminal_tree.update_tree()

    def closeEvent(self, ev) -> None:
        """
        Before closing check if some things about the mapping (fully mapped, no duplicates). Then close the application
        """
        all_mapped = True
        parameter_duplicates = {}
        for terminal_params in self.terminal_parameters.values():
            for param in terminal_params.values():
                if param is None:
                    all_mapped = False
                elif isinstance(param, Parameter):
                    param_hash = hash(param)
                    if param_hash in parameter_duplicates:
                        parameter_duplicates[param_hash].append(param)
                    else:
                        parameter_duplicates[param_hash] = [param]
                else:
                    raise TypeError("Gate parameters have to be either None or of type Parameter.")

        # Not every terminal parameter is mapped
        if not all_mapped:
            dialog = MessageBox_notallmapped(self)
            answer = dialog.exec()
            if not answer == QMessageBox.Yes:
                ev.ignore()
                return

        # Give warning if there are duplicates
        for items_with_param in parameter_duplicates.values():
            if len(items_with_param) != 1:
                dialog = MessageBox_duplicates(self)
                answer = dialog.exec()
                # question = ("Do you really want to stop the mapping process? "
                #             "Multiple terminal parameters are mapped to the same parameter!")
                # answer = dialog.question(self, "", question,  QMessageBox.Yes | QMessageBox.No)
                if not answer == QMessageBox.Yes:
                    ev.ignore()
                    return

        # properly close application
        self.terminal_tree.monitoring_timer.stop()
        QApplication.exit()
        return super().closeEvent(ev)

    def show_help(self):
        gui_help_txt_path = __file__.replace("mapping_gui.py", "GUI_help.txt")
        with open(gui_help_txt_path) as f:
            help_txt = f.read()

        self.help_window = ScrollLabel(help_txt)

        self.help_window.show()

    def monitoring_set_get_type(self):
        self.use_get_action.setChecked(True)
        self.use_cache_action.setChecked(False)
        self.terminal_tree.monitoring_get_type = "get"

    def monitoring_set_cache_type(self):
        self.use_cache_action.setChecked(True)
        self.use_get_action.setChecked(False)
        self.terminal_tree.monitoring_get_type = "cache"

    @pyqtSlot(QStandardItem, QStandardItem)
    def drag_terminal_drop_instr_slot(self, instr_elem, terminal_elem):
        """
        This receives signal from InstrumentTree if two elements were paired via drag and drop.
        Mapping is carried out from here (mainwindow)
        """
        self.map_given_terminal_instrument_elem_selection(terminal_elem.source, instr_elem.source)
        self.terminal_tree.update_tree()

    def toggle_monitoring(self):
        if self.monitoring_enable:
            self.terminal_tree.monitoring_timer.stop()
            self.monitoring_enable = False
            self.toggle_monitoring_action.setText("Enable")
        else:
            self.terminal_tree.monitoring_timer.start(1000)
            self.monitoring_enable = True
            self.toggle_monitoring_action.setText("Disable")

        self.terminal_tree.setColumnHidden(2, not self.monitoring_enable)

    def set_refresh_rate(self):
        """
        Set refresh rate of monitoring via input dialog.
        """
        val, ok = QInputDialog.getDouble(
            self, "Refresh delay [s]", "Refresh delay [s]", 1, 0.01, 100, 2, Qt.WindowFlags(), 0.1
        )
        if ok:
            self.terminal_tree.monitoring_timer.stop()
            self.terminal_tree.showColumn(2)
            self.toggle_monitoring_action.setText("Disable")
            self.terminal_tree.monitoring_timer.start(int(val * 1000))
            self.monitoring_enable = True

    def map_parameter(self, parameter: Parameter, traverse: tuple[str, str]):
        """
        Maps a instrument parameter to a specific terminal parameter accessed by the given traversal info.
        Doesn't do much anymore, but I kept this around for slightly better readability
        (and easier refactoring if necessary).
        """
        self.terminal_parameters[traverse[0]][traverse[1]] = parameter

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """
        Handles keyboard shortcuts and mapping using enter key. Selecting an instrument in the terminal_tree and
        pressing enter will switch focus to the instrument_tree and select a suitable mapping candidate.
        The user can change the selection and press enter again to do the mapping. The focus switches back to the
        terminal_tree and a new terminal is selected.
        """
        if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
            sel_idx = []
            for idx in self.terminal_tree.selectedIndexes():
                if idx.column() == 0:
                    sel_idx.append(idx)

            if self.terminal_tree.hasFocus() and len(sel_idx) == 1:
                # Element in terminal tree has been selected.

                # select element and switch focus to instruments
                self.terminal_tree.selected_terminal_tree_elem = sel_idx[0]
                selected_item = self.terminal_tree.model().itemFromIndex(self.terminal_tree.selected_terminal_tree_elem)
                self.instrument_tree.setFocus()

                # select instrument element based on predicted mapping
                if selected_item.hasChildren():
                    _terminal_params = get_children(selected_item)
                    terminal_params = []
                    for terminal_param in _terminal_params:
                        terminal_params.append(terminal_param.source[1])
                else:
                    terminal_params = [selected_item.source[1]]

                perfect_mappings = self.instrument_tree.get_perfect_mappings(terminal_params)

                # set idx to first perfect_mapping that has not yet been fully mapped to a terminal
                idx = self.instrument_tree.model().invisibleRootItem().child(0, 0).index()  # default
                for perfect_mapping in perfect_mappings:
                    instr_mapped = False
                    for terminal in get_children(self.terminal_tree.model().invisibleRootItem()):
                        if (
                            self.terminal_tree.model().data(terminal.index().siblingAtColumn(1))
                            == perfect_mapping.source.full_name
                        ):
                            instr_mapped = True
                            break

                    if not instr_mapped:
                        idx = perfect_mapping.index()
                        break

                self.instrument_tree.selectionModel().select(idx, QItemSelectionModel.SelectionFlag.ClearAndSelect)
                self.instrument_tree.setCurrentIndex(idx)
            elif (
                self.instrument_tree.hasFocus()
                and self.terminal_tree.selected_terminal_tree_elem is not None
                and len(self.instrument_tree.selectedIndexes()) == 1
            ):
                # Element in instrument tree selected. Start mapping
                terminal_elem = self.terminal_tree.model().itemFromIndex(self.terminal_tree.selected_terminal_tree_elem)

                idx_instr = self.instrument_tree.selectedIndexes()[0]
                instr_elem = self.instrument_tree.model().itemFromIndex(idx_instr)

                # Check if it is already mapped to display dialog
                any_mapped = False
                if type(terminal_elem.source[1]) is str:
                    terminal_params = (terminal_elem.source[1],)
                else:
                    terminal_params = terminal_elem.source[1]

                for terminal_param in terminal_params:
                    if not self.terminal_parameters[terminal_elem.source[0]][terminal_param] is None:
                        any_mapped = True
                        break

                    if any_mapped:
                        break

                if any_mapped:
                    dialog = MessageBox_overwrite(self)
                    answer = dialog.exec()
                    if answer == QMessageBox.Yes:
                        do_mapping = True
                    else:
                        do_mapping = False
                else:
                    do_mapping = True

                if do_mapping:
                    mapped = self.map_given_terminal_instrument_elem_selection(terminal_elem.source, instr_elem.source)
                    if mapped:
                        self.terminal_tree.selected_terminal_tree_elem = None

                        # select next not fully mapped terminal (quick navigation)
                        self.terminal_tree.setFocus()
                        for terminal_name, terminal in self.terminal_tree.terminal_parameters.items():
                            all_mapped = True
                            for terminal_param_name, terminal_param in terminal.items():
                                if terminal_param is None:
                                    all_mapped = False
                                    break

                            if not all_mapped:
                                terminal_elem = get_child(self.terminal_tree.model().invisibleRootItem(), terminal_name)

                                self.terminal_tree.selectionModel().select(
                                    terminal_elem.index(), QItemSelectionModel.SelectionFlag.ClearAndSelect
                                )
                                self.terminal_tree.setCurrentIndex(terminal_elem.index())
                                break
            # self.terminal_tree.update_tree()
        elif event.key() == Qt.Key_A:
            self.map_automatically()
        elif event.key() == Qt.Key_R:
            self.reset_mapping()
        elif event.key() == Qt.Key_E:
            self.close()
        elif event.key() == Qt.Key_U:
            self.unfold_terminals()

        self.terminal_tree.update_tree()

        return super().keyPressEvent(event)

    def map_given_terminal_instrument_elem_selection(
        self,
        terminal_tree_traversal: tuple[str, tuple[str]] | tuple[str, str],
        instr_elem: Metadatable | Parameter,
    ) -> bool:
        """
        For a selected item in terminal_tree (given via terminal_tree_traversal) and selected item
        in instrument_tree (instr_elem) do the mapping process.
        Behaviour based on combinations like: direct mapping between parameters, automap to all children etc.
        """
        tree = self.terminal_tree
        mapped = False
        if isinstance(terminal_tree_traversal[1], str) and isinstance(instr_elem, Parameter):
            # map directly
            self.map_parameter(instr_elem, terminal_tree_traversal)
            mapped = True
        elif isinstance(terminal_tree_traversal[1], tuple) and isinstance(instr_elem, (InstrumentModule, Instrument)):
            # map automatically as much as possible
            all_params = filter_flatten_parameters(instr_elem)
            child_params_all = {
                param_name: param
                for param_name, param in all_params.items()
                if instr_elem in param.instrument.ancestors
            }

            # try to map chosen terminal_parameters with child parameters of instr_elem if uniquely possible
            possible_mappings = get_possible_mapping_candidates(terminal_tree_traversal[1], child_params_all)
            for terminal_param_name, parameter_candidates in possible_mappings.items():
                if len(parameter_candidates) == 1:
                    # unique mapping - iterate over terminal_parameters of terminal

                    terminal = get_child(tree.model().invisibleRootItem(), terminal_tree_traversal[0])
                    terminal_param = get_child(terminal, terminal_param_name)
                    self.map_parameter(parameter_candidates[0], terminal_param.source)
                    mapped = True

        elif isinstance(terminal_tree_traversal[1], str) and isinstance(instr_elem, (InstrumentModule, Instrument)):
            # map automatically to one parameter
            all_params = filter_flatten_parameters(instr_elem)
            child_params_all = {
                param_name: param
                for param_name, param in all_params.items()
                if instr_elem in param.instrument.ancestors
            }
            possible_mappings = get_possible_mapping_candidates((terminal_tree_traversal[1],), child_params_all)

            # dict is always of size 1 (only one terminal_parameter)
            terminal_param_name = terminal_tree_traversal[1]
            possible_instrument_parameters = possible_mappings[terminal_param_name]
            if len(possible_instrument_parameters) == 1:
                if terminal_param_name == terminal_tree_traversal[1]:
                    self.map_parameter(possible_instrument_parameters[0], terminal_tree_traversal)
                    mapped = True
        elif isinstance(terminal_tree_traversal[1], tuple) and isinstance(instr_elem, Parameter):
            # map parameter of chosen terminal to chosen instrument parameter
            parameter = instr_elem
            terminal_parameters = terminal_tree_traversal[1]
            if hasattr(parameter, "_mapping"):
                if parameter._mapping in terminal_parameters:
                    self.map_parameter(parameter, (terminal_tree_traversal[0], parameter._mapping))
                    mapped = True

        return mapped

    def unfold_terminals(self):
        """
        Unfolds all terminals in the terminal_tree (quickly make everything visible). Collapse all if already unfolded
        """
        all_expanded = True
        for terminal in get_children(self.terminal_tree.model().invisibleRootItem()):
            if not self.terminal_tree.isExpanded(terminal.index()):
                all_expanded = False
                break

        if all_expanded:
            self.terminal_tree.collapseAll()
        else:
            self.terminal_tree.expandAll()

        self.terminal_tree.update_tree()

    def map_automatically(self):
        """
        Map all terminals automatically. The algorithm used is (almost) equivalent to selecting the first terminal and
        repeatedly pressing the enter key until the last terminal (in the tree) is mapped. This works best if the
        terminals are in the same order as the instruments that they should be mapped to. Additionally the terminals
        mapping to channels of an instrument should be ordered the same as the channels
        (up to the driver but usually something like 0,1,2,...).
        """
        self.reset_mapping()
        for terminal_name, terminal in self.terminal_parameters.items():
            _perfect_mappings = self.instrument_tree.get_perfect_mappings(terminal.keys())

            # filtering out parent channels (instrument) that would also lead to unique mapping
            # This is an edge case for instruments that have a single channel of some type
            # (then both the channel and the instrument are uniquely mappable)
            perfect_mappings = []
            for _perfect_mapping1 in _perfect_mappings:
                # check if there is a mapping which has _perfect_mapping1 as parent
                child_in_perfect_mappings = False
                for _perfect_mapping2 in _perfect_mappings:
                    if _perfect_mapping1.source is _perfect_mapping2.source.parent:
                        child_in_perfect_mappings = True
                        break

                if not child_in_perfect_mappings:
                    perfect_mappings.append(_perfect_mapping1)

            # map to first perfect_mapping that has not yet been fully mapped to a terminal
            for perfect_mapping in perfect_mappings:
                perfect_mapping_channel_name = perfect_mapping.source.full_name
                instr_mapped = False
                # find out if perfect mapping candidate has been already perfectly mapped to another terminal
                for _terminal_name, _terminal in self.terminal_parameters.items():
                    all_mapped_to_same_channel = True
                    for terminal_param_name, param in _terminal.items():
                        if param is not None:
                            if param.instrument.full_name != perfect_mapping_channel_name:
                                all_mapped_to_same_channel = False
                                break
                        else:
                            all_mapped_to_same_channel = False
                            break

                    if all_mapped_to_same_channel:
                        instr_mapped = True
                        break

                if not instr_mapped:
                    # map to channel
                    terminal_element = (terminal_name, tuple(terminal.keys()))
                    self.map_given_terminal_instrument_elem_selection(terminal_element, perfect_mapping.source)
                    break

        self.terminal_tree.update_tree()

    # Not used anymore. Unique, but algorithm to weak to be useful in practise
    def map_automatically_unique(self):
        """
        Automatically map all unique terminal_parameter instrument_parameter pairs. If there are multiple
        terminal_parameters with the same name their unique mapping is impossible.
        """
        # call get_possible_mapping_candidates for each terminal
        terminal_mapping_candidates = {}
        terminal_parameters_occurances = {}
        all_params = filter_flatten_parameters(self.components)
        for terminal_name, terminal_params in self.terminal_tree.terminal_parameters.items():
            terminal_mapping_candidates[terminal_name] = get_possible_mapping_candidates(
                terminal_params.keys(), all_params
            )
            for terminal_param in terminal_params.keys():
                if terminal_param in terminal_parameters_occurances.keys():
                    terminal_parameters_occurances[terminal_param] += 1
                else:
                    terminal_parameters_occurances[terminal_param] = 1

        # a terminal parameter can be mapped uniquely if its mapped parameter only appears once in the candidate
        # dictionaries of all terminals
        # For a unique mapping between terminal parameter and instrument parameter:
        #   1. a terminal parameter (name) must only occur a single time for all terminals
        #   2. unique mapping from terminal parameter to instrument_parameter (true if list of candidates has length 1)
        for terminal, terminal_params in self.terminal_tree.terminal_parameters.items():
            all_mapped = True
            for terminal_param in terminal_params.keys():
                if not terminal_parameters_occurances[terminal_param] == 1:
                    all_mapped = False
                    continue
                if not len(terminal_mapping_candidates[terminal][terminal_param]) == 1:
                    all_mapped = False
                    continue

                self.map_parameter(terminal_mapping_candidates[terminal][terminal_param][0], (terminal, terminal_param))

            if all_mapped:
                self.terminal_tree.collapse(get_child(self.terminal_tree.model().invisibleRootItem(), terminal).index())
            else:
                self.terminal_tree.expand(get_child(self.terminal_tree.model().invisibleRootItem(), terminal).index())

        self.terminal_tree.update_tree()

    def reset_mapping(self):
        """
        Reset all mappings. Reset dictionary which actually holds the mapping. Reset Tree representation.
        """
        for terminal_name, terminal_params in self.terminal_parameters.items():
            self.terminal_parameters[terminal_name] = {
                terminal_param_name: None for terminal_param_name in terminal_params.keys()
            }

        self.terminal_tree.selected_terminal_tree_elem = None
        self.terminal_tree.update_tree()
        self.terminal_tree.setFocus()


def get_possible_mapping_candidates(
    terminal_params: tuple[str], instrument_parameters: Mapping[Any, Parameter]
) -> Mapping[Any, list[Parameter]]:
    """
    For input terminal and collection of instrument_parameters: get dictionary with key:
    terminal parameter name, value: list(parameters that can be mapped to that terminal parameter)
    Similar to base.py _map_gate_to_instrument
    """
    mapped_parameters = {
        key: parameter for key, parameter in instrument_parameters.items() if hasattr(parameter, "_mapping")
    }
    mapping = {}
    for terminal_param in terminal_params:
        candidates = [parameter for parameter in mapped_parameters.values() if parameter._mapping == terminal_param]
        mapping[terminal_param] = candidates

    return mapping


def get_children(parent: QStandardItem) -> list[QStandardItem]:
    """Return list of children (QStandardItem) of parent"""
    children = []
    num_rows = parent.rowCount()
    if num_rows == 0:
        return []

    _row = 0
    while True:
        children.append(parent.child(_row, 0))
        if _row == num_rows - 1:
            break

        _row = _row + 1

    return children


def get_child(parent: QStandardItem, text: str) -> QStandardItem | None:
    for child in get_children(parent):
        if child.text() == text:
            return child

    return None


def traverse_tree(root: QStandardItem, traversal_names: list(str)) -> QStandardItem:
    def traverse(parent, names):
        if names == []:
            return parent

        child = get_child(parent, names.pop(0))
        if child is None:
            print("problem")
        else:
            return traverse(child, names)

    return traverse(root, traversal_names)


class ScrollLabel(QScrollArea):
    """
    Scrollable Label for displaying help window
    """

    def __init__(self, text: str):
        QScrollArea.__init__(self)

        self.setWidgetResizable(True)

        self.label = QLabel(self)
        self.setWidget(self.label)
        self.label.setIndent(10)
        self.label.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.label.setWordWrap(True)

        self.label.setText(text)

    def setText(self, text):
        self.label.setText(text)


class MessageBox_notallmapped(QMessageBox):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle("Warning! Not all parameters mapped!")
        self.setIcon(QMessageBox.Warning)
        self.setText("Do you really want to stop the mapping process? Not all parameters are mapped!")
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)


class MessageBox_duplicates(QMessageBox):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle("Warning! Duplicate mapping!")
        self.setIcon(QMessageBox.Warning)
        self.setText(
            "Do you really want to stop the mapping process? "
            "Multiple terminal parameters are mapped to the same parameter!"
        )
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)


class MessageBox_overwrite(QMessageBox):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle("Warning! Duplicate mapping!")
        self.setIcon(QMessageBox.Warning)
        self.setText("Do you really want to change/overwrite the existing mapping?")
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)


def map_terminals_gui(
    components: Mapping[Any, Metadatable],
    terminal_parameters: TerminalParameters,
    existing_terminal_parameters: TerminalParameters | None = None,
    metadata: Metadata | None = None,
    monitoring: bool = True,
    skip_gui_if_mapped: bool = True,
) -> None:
    """
    Maps the terminals, that were defined in the MeasurementScript to the instruments, that are initialized in QCoDeS.

    Args:
        components (Mapping[Any, Metadatable]): Instruments/Components in QCoDeS
        terminal_parameters (Mapping[Any, Union[Mapping[Any, Parameter], Parameter]]):
            Terminals, as defined in the measurement script
        existing_terminal_parameters (Mapping[Any, Union[Mapping[Any, Parameter], Parameter]] | None):
            Already existing mapping, that is used to automatically create the mapping
            for already known terminals without user input.
        metadata (Metadata | None): If provided, add mapping to the metadata object.
        monitoring (bool):
            If True the mapped parameters are periodically read out
            (either by get command (default) or cached value)
        skip_gui_if_mapped (bool):
            If True and existing_terminal_parameters completely covers all terminal_parameters,
            dont open gui and continue
    """
    if existing_terminal_parameters is None:
        # reset in case there is already some mapping
        for terminal_name, parameter_mapping in terminal_parameters.items():
            for terminal_parameter, instr_parameter in parameter_mapping.items():
                terminal_parameters[terminal_name][terminal_parameter] = None

        run_gui = True
    else:
        # try to get the mapping from an existing mapping (if the respective terminal name exists)
        for terminal_name, terminal_params in terminal_parameters.items():
            if terminal_name in existing_terminal_parameters:
                for terminal_param_name, mapped_param in terminal_params.items():
                    if terminal_param_name in existing_terminal_parameters[terminal_name]:
                        terminal_parameters[terminal_name][terminal_param_name] = existing_terminal_parameters[
                            terminal_name
                        ][terminal_param_name]

        all_mapped = True
        for terminal_params in terminal_parameters.values():
            for param in terminal_params.values():
                if param is None:
                    all_mapped = False
                    break

            if not all_mapped:
                break

        run_gui = not all_mapped

    if run_gui or not skip_gui_if_mapped:
        if QApplication.instance() is None:
            app = QApplication([])
        else:
            app = QApplication.instance()

        w = MainWindow(
            components,
            terminal_parameters,
            monitoring=monitoring,
        )

        w.show()
        app.exec_()

    # if metadata is provided, add mapping to metadata object
    if metadata is not None:
        j = json.dumps(terminal_parameters, default=lambda o: str(o))
        metadata.add_terminal_mapping(j, "custom-mapping")



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - 4K User
# - Daniel Grothe
# - Till Huckeman


from qcodes.instrument.parameter import Parameter

from qumada.instrument.custom_drivers.Harvard.Decadac import Decadac
from qumada.instrument.mapping import DECADAC_MAPPING
from qumada.instrument.mapping.base import InstrumentMapping


class DecadacMapping(InstrumentMapping):
    def __init__(self):
        super().__init__(DECADAC_MAPPING, is_triggerable=True)
        self._trigger_in: str | None = None
        self.AVAILABLE_TRIGGERS: list = ["trigger_in_1", "trigger_in_2"]

    def ramp(
        self,
        parameters: list[Parameter],
        *,
        start_values: list[float] | None = None,
        end_values: list[float],
        ramp_time: float,
        block: bool = False,
        sync_trigger=None,
        sync_trigger_level=2,
    ) -> None:
        assert len(parameters) == len(end_values)
        if start_values is not None:
            assert len(parameters) == len(start_values)

        if len(parameters) > 1:
            raise Exception("Maximum length of rampable parameters currently is 1.")
        # TODO: Test delay when ramping multiple parameters in parallel.
        # TODO: Add Trigger option?
        # check, if all parameters are from the same instrument
        instruments = {parameter.root_instrument for parameter in parameters}
        if len(instruments) > 1:
            raise Exception("Parameters are from more than one instrument. This would lead to non synchronized ramps.")

        instrument: Decadac = instruments.pop()
        assert isinstance(instrument, Decadac)

        if not start_values:
            start_values = [param.get() for param in parameters]
        # ramp_rates = np.abs((np.array(end_values) - np.array(start_values)) / np.array(ramp_time))
        # if sync_trigger:
        #     if sync_trigger in parameters:
        #         raise Exception("Synchronized trigger cannot be part of parameters")
        #     assert isinstance(sync_trigger.root_instrument, Decadac)
        #     sync_trigger._instrument.enable_ramp(False)
        #     sync_trigger.set(sync_trigger_level)
        for param, start_value, end_value, ramp_time in zip(
            parameters, start_values, end_values, [ramp_time for _ in parameters]
        ):
            param._instrument._script_ramp(start_value, end_value, ramp_time, trigger=self.trigger_mode)
        # if sync_trigger:
        #     sync_trigger.set(0)

    def trigger(self, parameter, level=1) -> None:
        instrument: Decadac = parameter.root_instrument
        assert isinstance(instrument, Decadac)
        parameter._instrument.enable_ramp(False)
        parameter.volt.set(level)

    def setup_trigger_in(self, trigger_settings: dict):
        # trigger_dict = {
        #     "always": 0,
        #     "trig1_low": 2,
        #     "trig2_low": 3,
        #     "until_trig1_rising": 4,
        #     "until_trig2_rising": 5,
        #     "until_trig1_falling": 6,
        #     "until_trig2_falling": 7,
        #     "never": 8,
        #     "trig1_high": 10,
        #     "trig2_high": 11,
        #     "after_trig1_rising": 12,
        #     "after_trig2_rising": 13,
        #     "after_trig1_falling": 14,
        #     "after_trig2_falling": 15,
        # }
        # TRIGGER_MODE_MAPPING: dict = {
        #     "continuous": 0,
        #     "edge": 1,
        #     "pulse": 3,
        #     "tracking_edge": 4,
        #     "tracking_pulse": 7,
        #     "digital": 6,
        # }
        print(
            "Warning: The Decadacs trigger level is fixed at roughly 1.69 V and cannot be changed. "
            "Please make sure that your triggers are setup accordingly"
        )
        trigger_mode = trigger_settings.get("trigger_mode", "continuous")
        polarity = trigger_settings.get("trigger_mode_polarity", "positive")

        if (trigger_mode, polarity) == ("edge", "positive"):
            mode = 12
        elif (trigger_mode, polarity) == ("edge", "negative"):
            mode = 14
        elif (trigger_mode, polarity) == ("digital", "positive"):
            mode = 10
        elif (trigger_mode, polarity) == ("digital", "negative"):
            mode = 2
        # TODO: Check other cases
        elif (trigger_mode, polarity) == ("continuous", "positive"):
            mode = 0
        elif (trigger_mode, polarity) == ("continuous", "negative"):
            mode = 0
        else:
            raise Exception("Selected trigger mode is not supported by DecaDac")

        if self.trigger_in is None:
            mode = 0
            print("No trigger input selected. Using continuous acquisition")
        if self.trigger_in == "trigger_in_2":
            mode += 1

        self.trigger_mode = mode

    @property
    def trigger_in(self):
        return self._trigger_in

    @trigger_in.setter
    def trigger_in(self, trigger: str | None) -> None:
        # TODO: Inform user about automatic changes of settings
        # TODO: This is done BEFORE the setup_buffer, so changes to trigger type will be overriden anyway?
        # print(f"Running trigger setter with: {trigger}")
        if trigger in self.AVAILABLE_TRIGGERS:
            self._trigger_in = trigger
        else:
            raise Exception(f"Trigger input {trigger} not available")
        if trigger is None:
            print("No Trigger provided!")



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Bertha Lab Setup
# - Daniel Grothe
# - Till Huckeman


from qcodes.parameters import Parameter
from qcodes_contrib_drivers.drivers.QDevil.QDAC1 import QDac

from qumada.instrument.mapping import QDAC_MAPPING
from qumada.instrument.mapping.base import InstrumentMapping


class QDacMapping(InstrumentMapping):
    def __init__(self):
        super().__init__(QDAC_MAPPING)

    def ramp(
        self,
        parameters: list[Parameter],
        *,
        start_values: list[float] | None = None,
        end_values: list[float],
        ramp_time: float,
        sync_trigger=None,
        **kwargs,
    ) -> None:
        assert len(parameters) == len(end_values)
        if start_values is not None:
            assert len(parameters) == len(start_values)

        if len(parameters) > 8:
            raise Exception("Maximum length of rampable parameters is 8.")

        # check, if all parameters are from the same instrument
        instruments = {parameter.root_instrument for parameter in parameters}
        if len(instruments) > 1:
            raise Exception("Parameters are from more than one instrument. This would lead to non synchronized ramps.")

        instrument: QDac = instruments.pop()
        assert isinstance(instrument, QDac)

        channellist = [instrument.channels.index(param._instrument) + 1 for param in parameters]

        if not start_values:
            start_values = []
        if sync_trigger is not None:
            parameters[0]._instrument.sync(sync_trigger)

        instrument.ramp_voltages(
            channellist=channellist,
            v_startlist=start_values,
            v_endlist=end_values,
            ramptime=ramp_time,
        )
        parameters[0]._instrument.sync(0)

    def setup_trigger_in():
        raise Exception("QDac does not have a trigger input!")



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckeman


import configparser


def load_from_config(section, key, config_file="../config.cfg"):
    """
    Helps you to load settings from the config file.
    """
    config = configparser.ConfigParser()
    config.read(config_file)
    if section in config:
        if key in config[section]:
            return config[section][key]
    print(f"No such entry in config: {section} {key}")
    return None


def save_to_config(section, key, value, config_file="../config.cfg"):
    """
    Stores settings in config file for you
    """
    config = configparser.ConfigParser()
    config.read(config_file)
    if section not in config:
        config[section] = {}
    config[section][key] = value
    with open(config_file, "w") as configfile:
        config.write(configfile)
    return None



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Bertha Lab Setup
# - Daniel Grothe
# - Max Lennart Oberländer
# - Sionludi Lab
# - Till Huckemann


from __future__ import annotations

import logging
import operator
import sys
import time
import warnings
from collections.abc import Sequence
from functools import partial
from typing import Any, Callable, Optional, Union

import matplotlib.axes
import matplotlib.colorbar
import numpy as np
from qcodes import config
from qcodes.dataset.data_set_protocol import DataSetProtocol
from qcodes.dataset.descriptions.detect_shapes import detect_shape_of_measurement
from qcodes.dataset.descriptions.versioning.rundescribertypes import Shapes
from qcodes.dataset.dond.do_nd_utils import (
    BreakConditionInterrupt,
    _handle_plotting,
    _register_actions,
    _register_parameters,
    _set_write_period,
)

try:
    from qcodes.dataset.dond.do_nd_utils import _catch_interrupts
except ImportError:
    from qcodes.dataset.dond.do_nd_utils import catch_interrupts as _catch_interrupts

from qcodes.dataset.experiment_container import Experiment
from qcodes.dataset.measurements import Measurement
from qcodes.dataset.threading import (  # SequentialParamsCaller,; ThreadPoolParamsCaller,
    process_params_meas,
)
from qcodes.parameters import ParameterBase
from tqdm.auto import tqdm

ActionsT = Sequence[Callable[[], None]]
BreakConditionT = Callable[[], bool]

ParamMeasT = Union[ParameterBase, Callable[[], None]]

AxesTuple = tuple[matplotlib.axes.Axes, matplotlib.colorbar.Colorbar]
AxesTupleList = tuple[list[matplotlib.axes.Axes], list[Optional[matplotlib.colorbar.Colorbar]]]
AxesTupleListWithDataSet = tuple[
    DataSetProtocol,
    list[matplotlib.axes.Axes],
    list[Optional[matplotlib.colorbar.Colorbar]],
]
MultiAxesTupleListWithDataSet = tuple[
    tuple[DataSetProtocol, ...],
    tuple[list[matplotlib.axes.Axes], ...],
    tuple[list[Optional[matplotlib.colorbar.Colorbar]], ...],
]

LOG = logging.getLogger(__name__)


def do1d_parallel(
    *param_meas: ParamMeasT,
    param_set: list[ParamMeasT],
    setpoints: np.array,
    delay: float,
    enter_actions: ActionsT = (),
    exit_actions: ActionsT = (),
    write_period: float | None = None,
    measurement_name: str = "",
    exp: Experiment | None = None,
    do_plot: bool | None = None,
    use_threads: bool | None = None,
    additional_setpoints: Sequence[ParamMeasT] = tuple(),
    show_progress: None | None = None,
    log_info: str | None = None,
    break_condition: BreakConditionT | None = None,
    backsweep_after_break: bool = False,
    wait_after_break: float = 0,
) -> AxesTupleListWithDataSet:
    """
    Performs a 1D scan of all ``param_set`` according to "setpoints" in parallel,
    measuring param_meas at each step. In case param_meas is
    an ArrayParameter this is effectively a 2d scan.

    Args:
        param_set: The QCoDeS parameter to sweep over
        setpoints: Array of setpoints for param_set
        delay: Delay after setting parameter before measurement is performed
        *param_meas: Parameter(s) to measure at each step or functions that
          will be called at each step. The function should take no arguments.
          The parameters and functions are called in the order they are
          supplied.
        enter_actions: A list of functions taking no arguments that will be
            called before the measurements start
        exit_actions: A list of functions taking no arguments that will be
            called after the measurements ends
        write_period: The time after which the data is actually written to the
            database.
        additional_setpoints: A list of setpoint parameters to be registered in
            the measurement but not scanned. Not supported right now.
        measurement_name: Name of the measurement. This will be passed down to
            the dataset produced by the measurement. If not given, a default
            value of 'results' is used for the dataset.
        exp: The experiment to use for this measurement.
        do_plot: should png and pdf versions of the images be saved after the
            run. If None the setting will be read from ``qcodesrc.json`
        use_threads: If True measurements from each instrument will be done on
            separate threads. If you are measuring from several instruments
            this may give a significant speedup.
        show_progress: should a progress bar be displayed during the
            measurement. If None the setting will be read from ``qcodesrc.json`
        log_info: Message that is logged during the measurement. If None a default
            message is used.
        backsweep_after_break: If true, after a break condition is fulfilled a
            reversed sweep starting from the measurement point at which the
            condition was fulfilled and stopping at the first setpoint of the
            measurement is performed.

    Returns:
        The QCoDeS dataset.
    """
    if do_plot is None:
        do_plot = config.dataset.dond_plot
    if show_progress is None:
        show_progress = config.dataset.dond_show_progress

    meas = Measurement(name=measurement_name, exp=exp)
    if log_info is not None:
        meas._extra_log_info = log_info
    else:
        meas._extra_log_info = "Using 'qcodes.utils.dataset.doNd.do1d'"

    all_setpoint_params = (param_set[0],) + tuple(s for s in additional_setpoints)

    measured_parameters = tuple(param for param in param_meas if isinstance(param, ParameterBase))
    measured_params = (*param_set[1:], *param_meas)
    setpoints_length = len(setpoints)
    if backsweep_after_break:
        setpoints_length *= 2
    try:
        loop_shape = tuple(1 for _ in additional_setpoints) + (setpoints_length,)
        shapes: Shapes = detect_shape_of_measurement(measured_params, loop_shape)
        print(loop_shape)
    except TypeError:
        LOG.exception(f"Could not detect shape of {measured_parameters} " f"falling back to unknown shape.")
        shapes = None

    _register_parameters(meas, all_setpoint_params, setpoints=None)
    _register_parameters(meas, measured_params, setpoints=all_setpoint_params, shapes=shapes)
    _set_write_period(meas, write_period)
    _register_actions(meas, enter_actions, exit_actions)

    original_delay = param_set[0].post_delay
    param_set[0].post_delay = delay

    if use_threads is None:
        use_threads = config.dataset.use_threads

    tracked_setpoints = list()
    # do1D enforces a simple relationship between measured parameters
    # and set parameters. For anything more complicated this should be
    # reimplemented from scratch

    with _catch_interrupts() as interrupted, meas.run() as datasaver:
        dataset = datasaver.dataset
        additional_setpoints_data = process_params_meas(additional_setpoints)

        # flush to prevent unflushed print's to visually interrupt tqdm bar
        # updates
        sys.stdout.flush()
        sys.stderr.flush()

        sweep_data = {}
        for channel in measured_params:
            sweep_data[channel]: list[float] = []

        for set_point in tqdm(setpoints, disable=not show_progress):
            for param in param_set:
                param.set(set_point)
            tracked_setpoints.append(set_point)
            time.sleep(delay)
            datasaver.add_result(
                (param_set[0], set_point),
                *process_params_meas(measured_params, use_threads=use_threads),
                *additional_setpoints_data,
            )

            for channel in measured_params:
                sweep_data[channel].append(channel.get())
            if callable(break_condition):
                if break_condition(sweep_data):
                    if backsweep_after_break:
                        tracked_setpoints.reverse()
                        time.sleep(wait_after_break)
                        for set_point in tqdm(tracked_setpoints, disable=not show_progress):
                            for param in param_set:
                                param.set(set_point)
                            datasaver.add_result(
                                (param_set[0], set_point),
                                *process_params_meas(measured_params, use_threads=use_threads),
                                *additional_setpoints_data,
                            )
                        break
                    else:
                        warnings.warn("Break condition was met.")
                        break

    param_set[0].post_delay = original_delay

    return _handle_plotting(dataset, do_plot, interrupted())


def do1d_parallel_asym(
    *param_meas: ParamMeasT,
    param_set: list[ParamMeasT],
    setpoints: list[np.array],
    delay: float,
    enter_actions: ActionsT = (),
    exit_actions: ActionsT = (),
    write_period: float | None = None,
    measurement_name: str = "",
    exp: Experiment | None = None,
    do_plot: bool | None = None,
    use_threads: bool | None = None,
    additional_setpoints: Sequence[ParamMeasT] = tuple(),
    show_progress: None | None = None,
    log_info: str | None = None,
    break_condition: BreakConditionT | None = None,
    backsweep_after_break: bool = False,
    wait_after_break: float = 0,
) -> AxesTupleListWithDataSet:
    """
    Performs a 1D scan of all ``param_set`` according to "setpoints" in parallel,
    measuring param_meas at each step. In case param_meas is
    an ArrayParameter this is effectively a 2d scan.

    Args:
        param_set: The QCoDeS parameter to sweep over
        setpoints: Array of setpoints for param_set
        delay: Delay after setting parameter before measurement is performed
        *param_meas: Parameter(s) to measure at each step or functions that
          will be called at each step. The function should take no arguments.
          The parameters and functions are called in the order they are
          supplied.
        enter_actions: A list of functions taking no arguments that will be
            called before the measurements start
        exit_actions: A list of functions taking no arguments that will be
            called after the measurements ends
        write_period: The time after which the data is actually written to the
            database.
        additional_setpoints: A list of setpoint parameters to be registered in
            the measurement but not scanned. Not supported right now.
        measurement_name: Name of the measurement. This will be passed down to
            the dataset produced by the measurement. If not given, a default
            value of 'results' is used for the dataset.
        exp: The experiment to use for this measurement.
        do_plot: should png and pdf versions of the images be saved after the
            run. If None the setting will be read from ``qcodesrc.json`
        use_threads: If True measurements from each instrument will be done on
            separate threads. If you are measuring from several instruments
            this may give a significant speedup.
        show_progress: should a progress bar be displayed during the
            measurement. If None the setting will be read from ``qcodesrc.json`
        log_info: Message that is logged during the measurement. If None a default
            message is used.
        backsweep_after_break: If true, after a break condition is fulfilled a
            reversed sweep starting from the measurement point at which the
            condition was fulfilled and stopping at the first setpoint of the
            measurement is performed.

    Returns:
        The QCoDeS dataset.
    """
    if do_plot is None:
        do_plot = config.dataset.dond_plot
    if show_progress is None:
        show_progress = config.dataset.dond_show_progress

    meas = Measurement(name=measurement_name, exp=exp)
    if log_info is not None:
        meas._extra_log_info = log_info
    else:
        meas._extra_log_info = "Using 'qcodes.utils.dataset.doNd.do1d'"

    all_setpoint_params = (*param_set,) + tuple(s for s in additional_setpoints)
    all_setpoint_params = [
        *param_set,
    ] + [s for s in additional_setpoints]
    if len(all_setpoint_params) != len(setpoints):
        raise NotImplementedError("Setpoints list length does not match number of dynamic parameters")

    for entry in setpoints:
        if len(entry) != len(setpoints[0]):
            raise NotImplementedError(
                "Setpoints for different parameters have different length. This is not yet supported"
            )
    measured_parameters = tuple(param for param in param_meas if isinstance(param, ParameterBase))
    measured_params = param_meas
    setpoints_length = len(setpoints[0])
    if backsweep_after_break:
        setpoints_length *= 2
    try:
        loop_shape = tuple(1 for _ in additional_setpoints) + (setpoints_length,)
        shapes: Shapes = detect_shape_of_measurement(measured_params, loop_shape)
        print(loop_shape)
    except TypeError:
        LOG.exception(f"Could not detect shape of {measured_parameters} " f"falling back to unknown shape.")
        shapes = None

    _register_parameters(meas, [all_setpoint_params[0]], setpoints=None)
    _register_parameters(meas, all_setpoint_params[1:], setpoints=(all_setpoint_params[0],), shapes=shapes)
    print(f"Measured parameters: {measured_parameters}")
    _register_parameters(meas, measured_params, setpoints=(all_setpoint_params[0],), shapes=shapes)
    _set_write_period(meas, write_period)
    _register_actions(meas, enter_actions, exit_actions)

    original_delay = param_set[0].post_delay
    param_set[0].post_delay = delay

    if use_threads is None:
        use_threads = config.dataset.use_threads

    tracked_setpoints = list([] for _ in param_set)
    # do1D enforces a simple relationship between measured parameters
    # and set parameters. For anything more complicated this should be
    # reimplemented from scratch
    with _catch_interrupts() as interrupted, meas.run() as datasaver:
        dataset = datasaver.dataset
        additional_setpoints_data = process_params_meas(additional_setpoints)

        # flush to prevent unflushed print's to visually interrupt tqdm bar
        # updates
        sys.stdout.flush()
        sys.stderr.flush()

        for j in range(len(setpoints[0])):
            datasaver_list = []
            for i in range(len(param_set)):
                param_set[i].set(setpoints[i][j])
                tracked_setpoints[i].append(setpoints[i][j])
                time.sleep(delay)
                datasaver_list.append((param_set[i], setpoints[i][j]))
            datasaver.add_result(
                *datasaver_list,
                *process_params_meas(measured_params, use_threads=use_threads),
                *additional_setpoints_data,
            )
            if callable(break_condition):
                if break_condition():
                    if backsweep_after_break:
                        # tracked_setpoints.reverse()
                        # need nested reverse?
                        print("Break condition was met. Starting backsweep!")
                        tracked_setpoints = [setpoints[::-1] for setpoints in tracked_setpoints]
                        time.sleep(wait_after_break)
                        for j in range(len(tracked_setpoints[0])):
                            datasaver_backward_list = []
                            for i, param in enumerate(param_set):
                                # tqdm might not work anymore as intended; need j instead of object?
                                # for set_point in tqdm(tracked_setpoints, disable=not show_progress):
                                param.set(tracked_setpoints[i][j])
                                datasaver_backward_list.append((param_set[i], tracked_setpoints[i][j]))
                            datasaver.add_result(
                                *datasaver_backward_list,
                                *process_params_meas(measured_params, use_threads=use_threads),
                                *additional_setpoints_data,
                            )
                        break
                    else:
                        warnings.warn("Break condition was met.")
                        print(
                            "Break condition was met. This meassage pops up alone \
                               if there is some issue with warning.warn"
                        )
                        raise BreakConditionInterrupt("Break condition was met.")

    param_set[0].post_delay = original_delay

    return _handle_plotting(dataset, do_plot, interrupted())


def _interpret_breaks(break_conditions: list, **kwargs) -> Callable[[], bool] | None:
    """
    Translates break conditions and returns callable to check them.

    Parameters
    ----------
    break_conditions : List of dictionaries containing:
            "channel": Gettable parameter to check
            "break_condition": String specifying the break condition.
                    Syntax:
                        Parameter to check: only "val" supported so far.
                        Comparator: "<",">" or "=="
                        Value: float
                    The parts have to be separated by blanks.

    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    Callable
        Function, that returns a boolean, True if break conditions are fulfilled.

    """

    def eval_binary_expr(op1: Any, oper: str, op2: Any) -> bool:
        # evaluates the string "op1 [operator] op2
        # supports <, > and == as operators
        ops = {
            ">": operator.gt,
            "<": operator.lt,
            "==": operator.eq,
        }
        # Why convert explicitly to float?
        # op1, op2 = float(op1), float(op2)
        return ops[oper](op1, op2)

    def f(cond, ops):
        return partial(eval_binary_expr, cond["channel"].get_latest(), ops[1], float(ops[2]))()

    def check_conditions(conditions: list[Callable[[], bool]]):
        for cond in conditions:
            if f(cond[0], cond[1]):
                return True
        return False

    conditions = []

    # Create break condition callables
    for cond in break_conditions:
        ops = cond["break_condition"].split(" ")
        if ops[0] != "val":
            raise NotImplementedError(
                'Only parameter values can be used for breaks in this version. Use "val" for the break condition.'
            )
        conditions.append([cond, ops])

    return partial(check_conditions, conditions) if conditions else None


def _dev_interpret_breaks(break_conditions: list, sweep_values: dict, **kwargs) -> Callable[[], bool] | None:
    """
    Translates break conditions and returns callable to check them.

    Parameters
    ----------
    break_conditions : List of dictionaries containing:
            "channel": Gettable parameter to check
            "break_condition": String specifying the break condition.
                    Syntax:
                        Parameter to check: only "val" supported so far.
                        Comparator: "<",">" or "=="
                        Value: float
                    The parts have to be separated by blanks.

    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    Callable
        Function, that returns a boolean, True if break conditions are fulfilled.

    """

    def return_false():
        return False

    def eval_binary_expr(op1: Any, oper: str, op2: Any) -> bool:
        # evaluates the string "op1 [operator] op2
        # supports <, > and == as operators
        ops = {
            ">": operator.gt,
            "<": operator.lt,
            "==": operator.eq,
        }
        # Why convert explicitly to float?
        # op1, op2 = float(op1), float(op2)
        return ops[oper](op1, op2)

    def check_conditions(conditions: list[Callable[[], bool]]):
        for cond in conditions:
            if cond():
                return True
        return False

    conditions = []
    # Create break condition callables
    for cond in break_conditions:
        ops = cond["break_condition"].split(" ")
        data = sweep_values[cond["channel"]]
        if ops[0] == "val":

            def f():
                return partial(eval_binary_expr, data[-1], ops[1], float(ops[2]))()

        elif ops[0] == "grad":
            if int(ops[1]) >= len(data):
                f = return_false
            elif float(ops[1]) < len(data):
                if data[len(data) - 1] != 0:
                    dx = (data[len(data) - 1] - data[len(data) - 1 - int(ops[1])]) / data[len(data) - 1]

                    def f():
                        return partial(eval_binary_expr, dx, ops[2], float(ops[3]))()

                else:
                    f = return_false
        else:
            raise NotImplementedError("NOT IMPLEMENTED")
        conditions.append(f)
    return check_conditions(conditions) if conditions else None



import logging
import multiprocessing as mp
import sys
import threading

from PyQt5.QtCore import (
    Q_ARG,
    QMetaObject,
    QObject,
    Qt,
    QThread,
    QTimer,
    pyqtSignal,
    pyqtSlot,
)
from PyQt5.QtWidgets import (
    QApplication,
    QLabel,
    QPushButton,
    QSpinBox,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

# Set up logging
logging.basicConfig(level=logging.WARNING, format="%(asctime)s - %(levelname)s - %(message)s")


class Parent:
    def __init__(self, name):
        self.name = name


class Parameter:
    def __init__(self, name, parent):
        self.name = name
        self._parent = parent

    def __call__(self):
        return 42  # Beispielwert

    def cached_value(self):
        return 42  # Beispielwert


class Worker(QObject):
    data_ready = pyqtSignal(list)

    def __init__(self, interval, parameters):
        super().__init__()
        self.interval = interval
        self.parameters = parameters
        self.running = True
        self.timer = QTimer()
        self.timer.timeout.connect(self.run)
        self.timer.start(self.interval)
        logging.debug("Worker initialized with interval %d ms", self.interval)

    def run(self):
        if self.running:
            data = [str(param()) for param in self.parameters]
            self.data_ready.emit(data)
            logging.debug("Worker emitted data: %s", data)
            QTimer.singleShot(self.interval, self.run)

    def stop(self):
        self.running = False
        self.timer.stop()
        logging.debug("Worker stopped")


class MeasurementGUI(QWidget):
    def __init__(self, parameters, data_queue):
        super().__init__()
        self.parameters = parameters
        self.data_queue = data_queue
        self.initUI()
        self.start_worker()

    def initUI(self):
        self.layout = QVBoxLayout()

        # Tabelle für Parameter
        self.table = QTableWidget(len(self.parameters), 2)
        self.table.setHorizontalHeaderLabels(["Parameter", "Value"])
        self.layout.addWidget(self.table)

        self.update_table()

        # Intervall-Einstellung
        self.interval_label = QLabel("Intervall (ms):")
        self.layout.addWidget(self.interval_label)

        self.interval_spinbox = QSpinBox()
        self.interval_spinbox.setRange(100, 10000)  # Bereich für das Intervall
        self.interval_spinbox.setValue(1000)  # Standardwert
        self.layout.addWidget(self.interval_spinbox)

        # Platzhalter für gecachte Werte
        self.cached_value_button = QPushButton("Gecachte Werte anzeigen")
        self.cached_value_button.clicked.connect(self.show_cached_values)
        self.layout.addWidget(self.cached_value_button)

        self.setLayout(self.layout)
        self.setWindowTitle("Messsoftware GUI")
        self.show()

        self.interval_spinbox.valueChanged.connect(self.update_interval)
        logging.debug("MeasurementGUI initialized")

    def start_worker(self):
        self.worker_thread = QThread()
        self.worker = Worker(self.interval_spinbox.value(), self.parameters)
        self.worker.moveToThread(self.worker_thread)
        self.worker.data_ready.connect(self.handle_data_ready)
        self.worker_thread.start()
        logging.debug("Worker thread started")

    def update_interval(self):
        self.worker.interval = self.interval_spinbox.value()
        logging.debug("Update interval to %d ms", self.interval_spinbox.value())

    @pyqtSlot(list)
    def handle_data_ready(self, data):
        logging.debug("Handling data ready in main thread: %s", data)
        QMetaObject.invokeMethod(self, "update_table_items", Qt.QueuedConnection, Q_ARG(list, data))

    @pyqtSlot(list)
    def update_table_items(self, data):
        logging.debug("Updating table items in main thread: %s", data)
        for row, value in enumerate(data):
            item = self.table.item(row, 1)
            if item is None:
                self.table.setItem(row, 1, QTableWidgetItem(value))
            else:
                item.setText(value)

    def update_table(self):
        logging.debug("Update table called")
        for row, param in enumerate(self.parameters):
            param_name = f"Param: {param._parent.name} {param.name}"
            self.table.setItem(row, 0, QTableWidgetItem(param_name))
            self.table.setItem(row, 1, QTableWidgetItem(str(param())))

    def show_cached_values(self):
        logging.debug("Show cached values called")
        cached_values = [str(param.cached_value()) for param in self.parameters]
        QMetaObject.invokeMethod(self, "update_table_items", Qt.QueuedConnection, Q_ARG(list, cached_values))

    def closeEvent(self, event):
        logging.debug("Close event called")
        self.worker.stop()
        self.worker_thread.quit()
        self.worker_thread.wait()
        event.accept()
        self.data_queue.put("QUIT")


def start_gui(parameters, data_queue):
    app = QApplication(sys.argv)
    ex = MeasurementGUI(parameters, data_queue)
    ex.show()
    logging.debug("GUI started")
    app.exec_()


def gui_process_main(parameters, data_queue):
    gui_process = mp.Process(target=start_gui, args=(parameters, data_queue))
    gui_process.start()
    return gui_process


def open_gui(parameters):
    data_queue = mp.Queue()
    gui_process = gui_process_main(parameters, data_queue)

    def data_listener():
        while True:
            try:
                data = data_queue.get()
                if data == "QUIT":
                    break
                print("Received data:", data)
            except KeyboardInterrupt:
                data_queue.put("QUIT")
                gui_process.join()
                break

    listener_thread = threading.Thread(target=data_listener, daemon=True)
    listener_thread.start()


if __name__ == "__main__":
    parent1 = Parent("Parent1")
    parent2 = Parent("Parent2")

    parameters = [
        Parameter("Parameter1", parent1),
        Parameter("Parameter2", parent1),
        Parameter("Parameter3", parent2),
    ]

    open_gui(parameters)



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - 4K User
# - Daniel Grothe
# - Jonas Mertens
# - Sionludi Lab
# - Till Huckeman


from .base import (
    MappingError,
    add_mapping_to_instrument,
    filter_flatten_parameters,
    map_gates_to_instruments,
)
from .mapping_gui import map_terminals_gui


def _build_path(subpath: str) -> str:
    """
    Build the path of the mapping file from this modules directory path and the subdirectory path.

    Args:
        subpath (str): Subpath of the JSON file.

    Returns:
        str: Path to the JSON file.
    """
    return __file__.replace("__init__.py", subpath)


DECADAC_MAPPING = _build_path("Harvard/Decadac.json")
SR830_MAPPING = _build_path("Stanford/SR830.json")
KEITHLEY_2400_MAPPING = _build_path("tektronix/Keithley_2400.json")
KEITHLEY_2450_MAPPING = _build_path("tektronix/Keithley_2450_voltage_source.json")
MFLI_MAPPING = _build_path("Zurich_Instruments/MFLI.json")
QDAC_MAPPING = _build_path("QDevil/QDac.json")
QDAC2_MAPPING = _build_path("QDevil/QDac2.json")
DUMMY_DMM_MAPPING = _build_path("Dummies/DummyDmm.json")
DUMMY_DAC_MAPPING = _build_path("Dummies/DummyDac.json")
DUMMY_CHANNEL_MAPPING = _build_path("Dummies/DummyChannel.json")

__all__ = [
    "MappingError",
    "add_mapping_to_instrument",
    "map_gates_to_instruments",
    "filter_flatten_parameters",
    "map_terminals_gui",
    "DECADAC_MAPPING",
    "SR830_MAPPING",
    "KEITHLEY_2400_MAPPING",
    "KEITHLEY_2450_MAPPING",
    "MFLI_MAPPING",
    "QDAC_MAPPING",
    "QDAC2_MAPPING",
    "DUMMY_DMM_MAPPING",
    "DUMMY_DAC_MAPPING",
]



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Sionludi Lab
# - Till Huckeman
# %%

import matplotlib
import matplotlib.ticker as ticker

# from qumada.instrument.mapping.base import flatten_list
import numpy as np
from matplotlib import pyplot as plt
from qcodes.dataset.data_export import reshape_2D_data

from qumada.utils.load_from_sqlite_db import (
    get_parameter_data,
    pick_measurements,
    separate_up_down,
)


# %%
def _handle_overload(*args, output_dimension: int = 1, x_name=None, y_name=None, z_name=None, **kwargs):
    """
    Reduces the amount of input parameters to output_dimension according to
    user input.
    """
    all_params = list(args)
    params = [None for i in range(output_dimension)]
    if len(all_params) == output_dimension:
        return all_params
    if x_name:
        for i, param in enumerate(all_params):
            if param[0] == x_name:
                params[0] = all_params.pop(i)
                output_dimension -= 1
    if y_name:
        for i, param in enumerate(all_params):
            if param[0] == y_name:
                params[1] = all_params.pop(i)
                output_dimension -= 1
    if z_name:
        for i, param in enumerate(all_params):
            if param[0] == z_name:
                params[2] = all_params.pop(i)
                output_dimension -= 1

    print(f"To many parameters found. Please choose {output_dimension} parameter(s)")
    for i in range(0, output_dimension):
        for idx, j in enumerate(all_params):
            print(f"{idx} : {j[0]}")
        choice = input("Please enter ID: ")
        for k in range(len(params)):
            if not params[k]:
                params[k] = all_params.pop(int(choice))
    return params


def _get_scaled_unit_and_factor(unit: str, values: list):
    """
    Determines the best scaling factor and prefix for the given values.
    """
    prefixes = {-12: "p", -9: "n", -6: "µ", -3: "m", 0: "", 3: "k", 6: "M", 9: "G"}
    abs_max_value = max(abs(min(values)), abs(max(values)))
    exponent = int(np.floor(np.log10(abs_max_value)) // 3 * 3) if abs_max_value != 0 else 0
    prefix = prefixes.get(exponent, "")
    scaling_factor = 10 ** (-exponent)
    return scaling_factor, f"{prefix}{unit}"


def _rescale_axis(axis, data, unit, axis_type="x"):
    """
    Rescales the axis ticks to avoid scientific notation and sets appropriate labels.
    """
    factor, scaled_unit = _get_scaled_unit_and_factor(unit, data)
    axis.set_major_formatter(ticker.FuncFormatter(lambda x, _: f"{x * factor:.0f}"))
    return factor, scaled_unit


def get_parameter_name_by_label(dataset, label):
    """
    Returns the parameter name for a given parameter label in a QCoDeS dataset.

    Parameters
    ----------
    dataset : qcodes.dataset
        The QCoDeS dataset containing the parameters.
    label : str
        The label of the parameter for which to find the name.

    Returns
    -------
    str
        The name of the parameter corresponding to the provided label.

    Raises
    ------
    ValueError
        If no parameter matches the label or if multiple parameters share the same label.
    """
    matching_parameters = []

    # Iterate over all parameters in the dataset and compare labels
    for parameter in dataset.get_parameters():
        if parameter.label == label:
            matching_parameters.append(parameter.name)

    # Handle different cases
    if not matching_parameters:
        raise ValueError(f"No parameter found with label '{label}'.")
    elif len(matching_parameters) > 1:
        raise ValueError(f"Multiple parameters found with label '{label}': {matching_parameters}")
    else:
        return matching_parameters[0]


def plot_2D(
    x_data,
    y_data,
    z_data,
    fig=None,
    ax=None,
    x_label=None,
    y_label=None,
    z_label=None,
    scale_axis=True,
    *args,
    **kwargs,
):
    """
    Plots 2D derivatives. Requires tuples of name and 1D arrays corresponding
    to x, y, and z data as input. Supports axis and colorbar scaling.

    Parameters
    ----------
    x_data, y_data, z_data : tuple
        Tuples containing name, values, units, and labels of x, y, and z axes.
    fig : matplotlib.figure.Figure, optional
        The figure object. Default is None.
    ax : matplotlib.axes.Axes, optional
        The axis object. Default is None.
    x_label, y_label, z_label : str, optional
        Custom axis labels. Default is None.
    scale_axis : bool, optional
        If True, rescales axes and colorbar to use SI prefixes. Default is True.
    *args, **kwargs
        Additional arguments for flexibility.

    Returns
    -------
    fig, ax : tuple
        The figure and axis objects for further customization.
    """
    if args:
        x_data, y_data, z_data = _handle_overload(x_data, y_data, z_data, *args, output_dimension=3)
    if ax is None or fig is None:
        fig, ax = plt.subplots()

    # Skalierung der Achsendaten und Einheiten
    x_values, y_values, z_values = x_data[1], y_data[1], z_data[1]
    x_unit, y_unit, z_unit = x_data[2], y_data[2], z_data[2]

    if scale_axis:
        x_scale, scaled_x_unit = _get_scaled_unit_and_factor(x_unit, x_values)
        y_scale, scaled_y_unit = _get_scaled_unit_and_factor(y_unit, y_values)
        z_scale, scaled_z_unit = _get_scaled_unit_and_factor(z_unit, z_values)

        # Skaliere die Daten
        x_values = np.array(x_values) * x_scale
        y_values = np.array(y_values) * y_scale
        z_values = np.array(z_values) * z_scale

        # Aktualisiere die Einheiten
        x_unit, y_unit, z_unit = scaled_x_unit, scaled_y_unit, scaled_z_unit

    # Reshape der Daten für 2D-Darstellung
    x, y, z = reshape_2D_data(x_values, y_values, z_values)

    # Plotten der 2D-Daten
    im = ax.pcolormesh(x, y, z, shading="auto")
    cbar = fig.colorbar(im, ax=ax)
    cbar.set_label(f"{z_data[3]} ({z_unit})")

    # Achsentitel aktualisieren
    x_label = x_label or f"{x_data[3]} ({x_unit})"
    y_label = y_label or f"{y_data[3]} ({y_unit})"

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    plt.tight_layout()
    plt.show()
    return fig, ax


# %%
def plot_2D_grad(x_data, y_data, z_data, *args, direction="both"):
    """
    Plots 2D derivatives. Requires tuples of name and 1D arrays corresponding
    to x, y and z data as input.
    Works well with QuMADA "get_parameter_data" method found in load_from_sqlite.
    direction argument can be x, y or z corresponding to the direction of the
    gradient used. "both" adds the gradients quadratically.

    TODO: Add get_parameter_data method as default to call when no data is provided
    TODO: Add further image manipulation and line detection functionality
    """
    if args:
        x_data, y_data, z_data = _handle_overload(x_data, y_data, z_data, *args, output_dimension=3)
    fig, ax = plt.subplots()
    x, y, z = reshape_2D_data(x_data[1], y_data[1], z_data[1])
    z_gradient = np.gradient(z, x, y)
    if direction == "both":
        grad = np.sqrt(z_gradient[0] ** 2 + z_gradient[1] ** 2)
    elif direction == "x":
        grad = z_gradient[0]
    elif direction == "y":
        grad = z_gradient[1]
    im = plt.pcolormesh(x, y, grad)
    fig.colorbar(im, ax=ax, label=f"Gradient of {z_data[0]} in {z_data[2]}/{x_data[2]}")
    plt.xlabel(f"{x_data[0]} in {x_data[2]}")
    plt.ylabel(f"{y_data[0]} in {y_data[2]}")
    plt.show()
    return fig, ax


# %%
def plot_2D_sec_derivative(x_data, y_data, z_data, *args):
    """
    Plots second derivative of data.
    Requires tuples of name and 1D arrays corresponding
    to x, y and z data as input.
    Works well with QuMADA "get_parameter_data" method found in load_from_sqlite.
    direction argument can be x, y or z corresponding to the direction of the
    gradient used. "both" adds the gradients quadratically.

    TODO: Add get_parameter_data method as default to call when no data is provided
    TODO: Add further image manipulation and line detection functionality
    """
    if args:
        x_data, y_data, z_data = _handle_overload(x_data, y_data, z_data, *args, output_dimension=3)
    fig, ax = plt.subplots()
    x, y, z = reshape_2D_data(x_data[1], y_data[1], z_data[1])
    z_gradient = np.gradient(z, x, y)
    z_g_x = np.gradient(z_gradient[0], x, y)
    z_g_y = np.gradient(z_gradient[1], x, y)
    grad2 = np.sqrt(z_g_x[0] ** 2 + z_g_y[1] ** 2 + 2 * z_g_x[1] * z_g_y[0])
    im = plt.pcolormesh(x, y, grad2)
    fig.colorbar(im, ax=ax, label="2nd Derivative of {z[1]}")
    plt.xlabel(f"{x_data[0]} in {x_data[2]}")
    plt.ylabel(f"{y_data[0]} in {y_data[2]}")
    plt.show()
    return fig, ax


# %%
def plot_hysteresis(dataset, x_name, y_name):
    fig, ax = plt.subplots()
    grad = np.gradient(dataset[x_name])
    curr_sign = np.sign(grad[0])
    data_list_x = list()
    data_list_y = list()
    start_helper = 0
    for i in range(0, len(grad)):
        if np.sign(grad[i]) != curr_sign:
            data_list_x.append(dataset[x_name][start_helper:i])
            data_list_y.append(dataset[y_name][start_helper:i])
            start_helper = i + 1
            curr_sign = np.sign(grad[i])
    data_list_x.append(dataset[x_name][start_helper : len(grad)])
    data_list_y.append(dataset[y_name][start_helper : len(grad)])

    for i in range(0, len(data_list_x)):
        plt.plot(data_list_x[i], data_list_y[i])
    plt.show()
    return fig, ax


# %%
def plot_hysteresis_new(x_data, y_data):
    fig, ax = plt.subplots()
    grad = np.gradient(x_data[1])
    curr_sign = np.sign(grad[0])
    signs = [curr_sign]
    data_list_x = list()
    data_list_y = list()
    start_helper = 0
    for i in range(0, len(grad)):
        if np.sign(grad[i]) != curr_sign:
            data_list_x.append(x_data[1][start_helper:i])
            data_list_y.append(y_data[1][start_helper:i])
            start_helper = i + 1
            curr_sign = np.sign(grad[i])
            signs.append(curr_sign)
    data_list_x.append(x_data[1][start_helper : len(grad)])
    data_list_y.append(y_data[1][start_helper : len(grad)])

    for i in range(0, len(data_list_x)):
        options = {-1: "v", 1: "^"}
        plt.plot(data_list_x[i], data_list_y[i], options[signs[i]])
    plt.show()
    return fig, ax


# %%


def plot_multiple_datasets(
    datasets: list = None,
    x_axis_parameters_name: str = None,
    y_axis_parameters_name: str = None,
    plot_hysteresis: bool = True,
    ax=None,
    fig=None,
    scale_axis=True,
    legend=True,
    exclude_string_from_legend: list = [],
    **kwargs,
):
    """
    Plot multiple datasets from a QCoDeS database into a single figure.

    This function supports plotting and can handle multiple datasets.
    It automatically manages axis labels, legends, and optionally rescales the axes
    to use appropriate SI prefixes (e.g., µA, mV) instead of scientific notation.

    Parameters
    ----------
    datasets : list, optional
        List of QCoDeS datasets to plot. If None, the function allows you to pick
        measurements from the currently loaded QCoDeS database. Default is None.
    x_axis_parameters_name : str, optional
        The name of the parameter to use for the x-axis. If None, you will be prompted
        to select it individually for each dataset if more than one parameter exists.
        Default is None.
    y_axis_parameters_name : str, optional
        The name of the parameter to use for the y-axis. If None, you will be prompted
        to select it individually for each dataset if more than one parameter exists.
        Default is None.
    plot_hysteresis : bool, optional
        If True, separates datasets with multiple sweeps into different curves based
        on the monotonicity of the x-axis data. For example, foresweep and backsweep
        can be plotted with distinct markers. Default is True.
    ax : matplotlib.axes._axes.Axes, optional
        Matplotlib axis to plot on. If None, a new figure and axis will be created.
        Default is None.
    fig : matplotlib.figure.Figure, optional
        Matplotlib figure object. Required if `ax` is provided. Default is None.
    scale_axis : bool, optional
        If True, rescales the x- and y-axes to use SI prefixes (e.g., µ, m, k) instead
        of scientific notation for better readability. Default is True.
    **kwargs : dict
        Additional keyword arguments for customizing the plot. For example:
            - font: int, font size for the plot.
            - marker: str, marker style for the data points.
            - markersize: int, size of the markers.
            - legend_fontsize: int, font size for the legend.
            - legend_markerscale: float, scale factor for legend markers.

    Returns
    -------
    ax : matplotlib.axes._axes.Axes
        The axis object containing the plotted data.

    Notes
    -----
    - This function assumes the input datasets are from QCoDeS and compatible with
      the `get_parameter_data` function.
    - Axis scaling is applied only when `scale_axis` is True, and the scaling factor
      is calculated based on the data range.
    - Monotonicity of the x-axis is used to detect and separate hysteresis loops.

    """

    if not datasets:
        datasets = pick_measurements()
    x_data = list()
    y_data = list()
    x_units = list()
    y_units = list()
    if kwargs.get("font", None) is not None:
        matplotlib.rc("font", size=35)
    if ax is None or fig is None:
        fig, ax = plt.subplots(figsize=(30, 30))
    x_labels = []
    y_labels = []
    for i in range(len(datasets)):
        label = datasets[i].name
        for string in exclude_string_from_legend:
            label = label.strip(string)
        x, y = _handle_overload(
            *get_parameter_data(datasets[i], y_axis_parameters_name),
            x_name=x_axis_parameters_name,
            y_name=y_axis_parameters_name,
            output_dimension=2,
        )
        x_data.append(x[1])
        y_data.append(y[1])
        x_labels.append(x[3])
        y_labels.append(y[3])
        x_units.append(x[2])
        y_units.append(y[2])

        if plot_hysteresis:
            x_s, y_s, signs = separate_up_down(x_data[i], y_data[i])
            for j in range(len(x_s)):
                if signs[j] == 1:
                    marker = "^"
                    f_label = f"{label} foresweep"
                else:
                    marker = "v"
                    f_label = f"{label} backsweep"
                plt.plot(
                    x_s[j],
                    y_s[j],
                    marker,
                    label=f_label,
                    markersize=kwargs.get("markersize", 15),
                )
        else:
            plt.plot(
                x_data[i],
                y_data[i],
                marker=kwargs.get("marker", "."),
                label=label,
                markersize=kwargs.get("markersize", 15),
            )

    # Scale axes and update labels
    if scale_axis is True:
        x_scaling_factor, x_units[0] = _rescale_axis(ax.xaxis, np.concatenate(x_data), x_units[0], "x")
        y_scaling_factor, y_units[0] = _rescale_axis(ax.yaxis, np.concatenate(y_data), y_units[0], "y")
    plt.xlabel(f"{x_labels[0]} ({x_units[0]})")
    plt.ylabel(f"{y_labels[0]} ({y_units[0]})")
    # Update x and y labels
    if legend is True:
        plt.legend(
            loc=kwargs.get("legend_position", "upper left"),
            fontsize=kwargs.get("legend_fontsize", 15),
            markerscale=kwargs.get("legend_markerscale", 1),
        )
    plt.tight_layout()

    return ax


# %%


def separate_up_down_old(x_data, y_data):
    grad = np.gradient(x_data)
    curr_sign = np.sign(grad[0])
    data_list_x = list()
    data_list_y = list()
    start_helper = 0
    for i in range(0, len(grad)):
        if np.sign(grad[i]) != curr_sign:
            data_list_x.append(x_data[start_helper:i])
            data_list_y.append(y_data[start_helper:i])
            start_helper = i + 1
            curr_sign = np.sign(grad[i])
    data_list_x.append(x_data[start_helper : len(grad)])
    data_list_y.append(y_data[start_helper : len(grad)])

    return data_list_x, data_list_y


# %%
def hysteresis(dataset, I_threshold=15e-12, parameter_name="lockin_current"):
    data = list(get_parameter_data(dataset=dataset, parameter_name=parameter_name))
    voltage = data[0][1]
    current = data[1][1]
    v, c = separate_up_down_old(voltage, current)
    for i in range(0, len(v[0])):
        if c[0][i] >= I_threshold:
            V_threshold_up = v[0][i]
            break
    for i in range(0, len(v[1])):
        if c[1][i] <= I_threshold:
            V_threshold_down = v[1][i]
            break
    return V_threshold_up, V_threshold_down



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe


from __future__ import annotations

from datetime import datetime
from typing import Protocol


class Metadata(Protocol):
    """
    Protocol for a Metadata object.
    Defines the methods to handle specific metadata portions collected with QuMADA.
    """

    def add_terminal_mapping(self, mapping: str, name: str):
        """Adds metadata of the mapping between terminals and instrument parameters as a (JSON) string."""

    def add_script_to_metadata(self, script: str, language: str, name: str):
        """Adds metadata of the used measurement script."""

    def add_parameters_to_metadata(self, parameters: str, name: str):
        """Adds parameters and their settings to metadata."""

    def add_datetime_to_metadata(self, dt: datetime):
        """Adds datetime to metadata."""

    def add_data_to_metadata(self, location: str, datatype: str, name: str):
        """Adds metadata related to the measurement data to metadata."""


class Savable(Protocol):
    """Protocol for a savable / updatable (metadata) object."""

    def save(self):
        """Saves metadata to new objects."""

    def update(self):
        """Updates the already saved metadata."""


class BasicMetadata:
    """Example implementation of a MetadataHandler."""

    def __init__(self):
        self.metadata = {}

    def add_terminal_mapping(self, mapping: str, name: str = "mapping"):
        self.metadata[name] = mapping

    def add_script_to_metadata(self, script: str, language: str, name: str = "script"):
        self.metadata[name] = (language, script)

    def add_parameters_to_metadata(self, parameters: str, name: str = "parameters"):
        self.metadata[name] = parameters

    def add_datetime_to_metadata(self, dt: datetime):
        self.metadata["datetime"] = dt

    def add_data_to_metadata(self, location: str, datatype: str, name: str = "data"):
        self.metadata[name] = (datatype, location)



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Till Huckeman
# - Daniel Grothe
# - Jonas Mertens
# - Sionludi Lab

from __future__ import annotations

import copy
import inspect
import json
import logging
import os
from abc import ABC, abstractmethod
from collections.abc import MutableSequence
from contextlib import suppress
from datetime import datetime
from functools import wraps
from typing import Any, Callable

import numpy as np
import qcodes as qc
from qcodes import Station
from qcodes.dataset import AbstractSweep, LinSweep
from qcodes.dataset.dond.do_nd_utils import ActionsT
from qcodes.parameters import Parameter, ParameterBase

from qumada.instrument.buffers import is_bufferable, is_triggerable
from qumada.metadata import Metadata
from qumada.utils.ramp_parameter import ramp_or_set_parameter
from qumada.utils.utils import flatten_array

logger = logging.getLogger(__name__)


def load_param_whitelist(folder_path=None):
    combined_params = set()
    if folder_path is None:
        folder_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../instrument/parameter_whitelists"))
        if not os.path.exists(folder_path):
            raise FileNotFoundError(f"Whitelist folder not found: {folder_path}")
    for filename in os.listdir(folder_path):
        if filename.endswith(".json"):  # Load all json
            file_path = os.path.join(folder_path, filename)
            with open(file_path) as file:
                data = json.load(file)
                for key in data.keys():
                    combined_params.update(data.get(key, []))
    return combined_params


def is_measurement_script(o):
    return inspect.isclass(o) and issubclass(o, MeasurementScript)


class QtoolsStation(Station):
    """Station object, inherits from qcodes Station."""


def create_hook(func, hook):
    """
    Decorator to hook a function onto an existing function.
    The hook function can use keyword-only arguments, which are omitted prior to execution of the main function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        hook(*args, **kwargs)
        # remove arguments used in hook from kwargs
        sig = inspect.signature(hook)
        varkw = next(
            filter(
                lambda p: p.kind is inspect.Parameter.VAR_KEYWORD,
                sig.parameters.values(),
            )
        ).name
        unused_kwargs = sig.bind(*args, **kwargs).arguments.get(varkw) or {}
        return func(*args, **unused_kwargs)

    return wrapper


class MeasurementScript(ABC):
    """
    Base class for measurement scripts.

    The abstract function "run" has to be implemented.
    """

    PARAMETER_NAMES: set[str] = load_param_whitelist()

    def __init__(self):
        # Create function hooks for metadata
        # reverse order, so insert metadata is run second
        self.run = create_hook(self.run, self._insert_metadata_into_db)
        self.run = create_hook(self.run, self._add_data_to_metadata)
        self.run = create_hook(self.run, self._add_current_datetime_to_metadata)

        self.properties: dict[Any, Any] = {}
        self.gate_parameters: dict[Any, dict[Any, Parameter | None] | Parameter | None] = {}
        self._buffered_num_points: int | None = None

    def add_gate_parameter(self, parameter_name: str, gate_name: str = None, parameter: Parameter = None) -> None:
        """
        Adds a gate parameter to self.gate_parameters.

        Args:
            parameter_name (str): Name of the parameter. Has to be in MeasurementScript.PARAMETER_NAMES.
            gate_name (str): Name of the parameter's gate. Set this, if you want to define the parameter
                             under a specific gate. Defaults to None.
            parameter (Parameter): Custom parameter. Set this, if you want to set a custom parameter. Defaults to None.
        """
        if parameter_name not in MeasurementScript.PARAMETER_NAMES:
            raise NameError(
                f'parameter_name "{parameter_name}" not in MeasurementScript.PARAMETER_NAMES. \
                            Allowed parameters are listed in qumada.instrument.parameter_whitelists'
            )
        if not gate_name:
            self.gate_parameters[parameter_name] = parameter
        else:
            # Create gate dict if not existing
            gate = self.gate_parameters.setdefault(gate_name, {})
            # Raise Exception, if gate "gate_name" was populated with a parameter (or smth. else) before
            if isinstance(gate, dict):
                gate[parameter_name] = parameter
            else:
                raise Exception("Gate {gate_name} is not a dictionary.")

    def _set_buffered_num_points(self) -> None:
        """
        Calculates number of datapoints when buffered measurements are performed and sets
        the buffered_num_points accordingly. Required to define QCoDeS datastructure.
        """

        if "burst_duration" in self.buffer_settings:
            self._burst_duration = float(self.buffer_settings["burst_duration"])

        if "duration" in self.buffer_settings:
            if "burst_duration" in self.buffer_settings:
                self._num_bursts = np.ceil(float(self.buffer_settings["duration"]) / self._burst_duration)
            elif "num_bursts" in self.buffer_settings:
                self._num_bursts = int(self.buffer_settings["num_bursts"])
                self._burst_duration = float(self.buffer_settings["duration"]) / self._num_bursts

        if "num_points" in self.buffer_settings:
            self.buffered_num_points = int(self.buffer_settings["num_points"])
            if "sampling_rate" in self.buffer_settings:
                self._burst_duration = float(self.buffered_num_points / self.buffer_settings["sampling_rate"])

        elif "sampling_rate" in self.buffer_settings:
            self._sampling_rate = float(self.buffer_settings["sampling_rate"])
            if self._burst_duration is not None:
                self.buffered_num_points = np.ceil(self._sampling_rate * self._burst_duration)
            elif all(k in self.buffer_settings for k in ("duration", "num_bursts")):
                self._burst_duration = float(self.buffer_settings["duration"] / self.buffer_settings["num_bursts"])

    def setup(
        self,
        parameters: dict,
        metadata: Metadata,
        *,
        add_script_to_metadata: bool = True,
        add_parameters_to_metadata: bool = True,
        buffer_settings: dict | None = None,
        measurement_name: str | None = None,
        **settings: dict,
    ) -> None:
        """
        Adds all gate_parameters that are defined in the parameters argument to
        the measurement. Allows to pass metadata to measurement and update the
        metadata with the script.

        Args:
            parameters (dict): Dictionary containing parameters and their settings
            metadata (Metadata): Object containing/handling metadata that should be
                                        available for the measurement.
            add_script_to_metadata (bool): If True (default), adds this object's content
                                           to the metadata.
            add_parameters_to_metadata (bool): If True (default), add the parameters to
                                               the metadata.
            settings (dict): Settings regarding the measurement script. Kwargs:
                ramp_rate: Defines how fast parameters are ramped during
                initialization and reset.
                setpoint_intervalle: Defines how smooth parameters are ramped
                during initialization and reset.
        """
        # TODO: Add settings to metadata
        self.metadata = metadata
        self.buffered = False
        self._lists_created = False
        self.measurement_name = measurement_name
        cls = type(self)
        try:
            self.buffer_settings.update(buffer_settings or {})
        except Exception as e:
            logger.warning(f"Ex={e}")
            self.buffer_settings = buffer_settings or {}
        self._set_buffered_num_points()

        try:
            self.settings.update(settings)
        except Exception as e:
            logger.warning(f"Ex={e}")
            self.settings = settings

        # Add script and parameters to metadata
        if add_script_to_metadata:
            try:
                metadata.add_script_to_metadata(inspect.getsource(cls), language="python", name=cls.__name__)
            except OSError as err:
                logger.warning(f"Source of MeasurementScript could not be acquired: {err}")
            except Exception as ex:
                logger.warning(f"Script could not be added to metadata: {ex}")

        if add_parameters_to_metadata:
            try:
                metadata.add_parameters_to_metadata(json.dumps(parameters), name=f"{cls.__name__}Settings")
            except Exception as ex:
                logger.warning(f"Parameters could not be added to metadata: {ex}")

        # Add gate parameters
        for gate, vals in parameters.items():
            self.properties[gate] = vals
            for parameter, properties in vals.items():
                self.add_gate_parameter(parameter, gate)

    def generate_lists(self) -> None:
        """
        Creates lists containing the corresponding parameters for further use.

        The .channels list always contain the QCoDes parameters that can for
        example directly be called to get the corresponding values.

        E.g.: ``[param() for param in self.gettable_channels]`` will return a list
        of the current values of all gettable parameters.

        The .parameters lists contain dictionaries with the keywords "gate" for the
        corresponding terminal name and "parameter" for the parameter name.
        This is usefull to get the keys for specific parameters from
        the gate_parameters.

        gettable and static lists both include static gettable parameters,
        the static_gettable lists only the ones that are both, static and
        gettable. This is e.g. useful for logging static parameters that cannot
        be buffered and thus cause errors in buffered measurements.

        """
        self.gettable_parameters: list[str] = []
        self.gettable_channels: list[str] = []
        self.static_gettable_parameters: list[str] = []
        self.static_gettable_channels: list[str] = []
        self.break_conditions: list[str] = []
        self.static_parameters: list[str] = []
        self.static_channels: list[str] = []
        self.dynamic_parameters: list[str] = []
        self.dynamic_channels: list[str] = []
        self.dynamic_sweeps: list[str] = []
        self.compensating_parameters: list[str] = []
        self.compensating_parameters_values: list[float] = []
        self.compensating_channels: list[str] = []
        self.compensating_leverarms: list[list[float]] = []
        self.compensated_parameters: list[list[str]] = []
        self.compensating_limits: list[list] = []
        self.abstract_parameters: list[str] = []
        self.abstract_setpoints: list[list] = []
        self.groups: dict[dict] = {}
        self.buffers: set = set()  # All buffers of gettable parameters
        self.trigger_ins: set = set()  # All trigger inputs that do not belong to buffers
        self.priorities: dict = {}
        self.loop: int = 0  # For usage with looped measurements

        for gate, parameters in self.gate_parameters.items():
            for parameter, channel in parameters.items():
                if gate == "abstract":
                    self.abstract_parameters.append()
                    self.abstract_setpoints.append(self.properties[gate][parameter]["setpoints"])

                if self.properties[gate][parameter]["type"].find("static") >= 0:
                    self.static_parameters.append({"gate": gate, "parameter": parameter})
                    self.static_channels.append(channel)
                    if self.properties[gate][parameter]["type"].find("gettable") >= 0:
                        self.static_gettable_parameters.append({"gate": gate, "parameter": parameter})
                        self.static_gettable_channels.append(channel)
                if self.properties[gate][parameter]["type"].find("gettable") >= 0:
                    self.gettable_parameters.append({"gate": gate, "parameter": parameter})
                    self.gettable_channels.append(channel)
                    with suppress(KeyError):
                        if self.properties[gate][parameter]["break_conditions"] is not None:
                            for condition in self.properties[gate][parameter]["break_conditions"]:
                                self.break_conditions.append({"channel": channel, "break_condition": condition})
                if self.properties[gate][parameter]["type"].find("comp") >= 0:
                    self.compensating_parameters.append({"gate": gate, "parameter": parameter})
                    self.compensating_channels.append(channel)
                    try:
                        self.compensating_parameters_values.append(self.properties[gate][parameter]["value"])
                    except KeyError as e:
                        print(
                            f"No value assigned for compensating parameter \
                              {self.compensating_parameters[-1]}"
                        )
                        raise e
                    try:
                        leverarms = self.properties[gate][parameter]["leverarms"]
                        assert isinstance(leverarms, list)
                        self.compensating_leverarms.append(self.properties[gate][parameter]["leverarms"])
                    except KeyError as e:
                        print(f"No leverarm specified for parameters {self.compensating_parameters[-1]}!")
                        raise e
                    try:
                        comp_list = []
                        for entry in self.properties[gate][parameter]["compensated_gates"]:
                            assert isinstance(entry, dict)
                            comp_list.append({"gate": entry["terminal"], "parameter": entry["parameter"]})
                        self.compensated_parameters.append(comp_list)
                    except KeyError as e:
                        print(
                            f"The terminal to be compensated for with {self.compensating_parameters[-1]} \
                            is not properly specified! Make sure to define a dictionary with \
                            terminal and parameter as keys."
                        )
                        raise e
                    try:
                        limits = self.properties[gate][parameter]["limits"]
                        self.compensating_limits.append(limits)
                    except KeyError as e:
                        print(
                            f"No limits assigned to compensating parameter \
                              {self.compensating_parameters[-1]}!"
                        )
                        raise e

                elif self.properties[gate][parameter]["type"].find("dynamic") >= 0:
                    self.dynamic_parameters.append({"gate": gate, "parameter": parameter})
                    self.dynamic_channels.append(channel)
                    if self.properties[gate][parameter].get("_is_triggered", False) and self.buffered:
                        if "num_points" in self.properties[gate][parameter].keys():
                            try:
                                assert self.properties[gate][parameter]["num_points"] == self.buffered_num_points
                            except AssertionError:
                                logger.warning(
                                    f"Number of datapoints from buffer_settings\
                                    and gate_parameters do not match. Using \
                                    the value from the buffer settings: \
                                    {self.buffered_num_points}"
                                )
                        elif "setpoints" in self.properties[gate][parameter].keys():
                            try:
                                assert len(self.properties[gate][parameter]["setpoints"]) == self.buffered_num_points
                            except AssertionError:
                                logger.warning(
                                    f"Number of datapoints from buffer_settings\
                                    and gate_parameters do not match. Using \
                                    the value from the buffer settings: \
                                    {self.buffered_num_points}"
                                )

                        else:
                            logger.info(
                                "No num_points or setpoints given for\
                                         buffered measurement. The value from \
                                         buffer_settings is used"
                            )
                        try:
                            self.dynamic_sweeps.append(
                                LinSweep(
                                    channel,
                                    self.properties[gate][parameter]["start"],
                                    self.properties[gate][parameter]["stop"],
                                    int(self.buffered_num_points),
                                    delay=self.properties[gate][parameter].setdefault("delay", 0),
                                )
                            )
                        except KeyError:
                            self.dynamic_sweeps.append(
                                LinSweep(
                                    channel,
                                    self.properties[gate][parameter]["setpoints"][0],
                                    self.properties[gate][parameter]["setpoints"][-1],
                                    int(self.buffered_num_points),
                                    delay=self.properties[gate][parameter].setdefault("delay", 0),
                                )
                            )
                    else:
                        try:
                            self.dynamic_sweeps.append(
                                LinSweep(
                                    channel,
                                    self.properties[gate][parameter]["start"],
                                    self.properties[gate][parameter]["stop"],
                                    int(self.properties[gate][parameter]["num_points"]),
                                    delay=self.properties[gate][parameter].setdefault("delay", 0),
                                )
                            )
                        except KeyError:
                            self.dynamic_sweeps.append(
                                CustomSweep(
                                    channel,
                                    self.properties[gate][parameter]["setpoints"],
                                    delay=self.properties[gate][parameter].setdefault("delay", 0),
                                )
                            )

                    # Only executed for dynamic parameters!
                    if "group" in self.properties[gate][parameter].keys():
                        group = self.properties[gate][parameter]["group"]
                        if group not in self.groups.keys():
                            self.groups[group] = {"channels": [], "parameters": [], "priority": None}
                        self.groups[group]["channels"].append(channel)
                        self.groups[group]["parameters"].append({"gate": gate, "parameter": parameter})
                        if self.groups[group]["priority"] is None:
                            if "priority" in self.properties[gate][parameter].keys():
                                if self.groups[group]["priority"] in self.priorities.keys():
                                    raise Exception("Assigned the same priority to multiple groups")
                                elif self.groups[group]["priority"] is None:
                                    self.groups[group]["priority"] = int(self.properties[gate][parameter]["priority"])
                                    self.priorities[int(self.groups[group]["priority"])] = self.groups[group]
                                    self.dynamic_parameters[-1]["priority"] = int(self.groups[group]["priority"])
                            else:
                                try:
                                    prio = int(group)
                                    if prio not in self.priorities.keys():
                                        self.groups[group]["priority"] = prio
                                        self.priorities[prio] = self.groups[group]
                                        self.dynamic_parameters[-1]["priority"] = prio
                                except ValueError:
                                    pass
                                except TypeError:
                                    pass

        if self.buffered:
            self.buffers = {
                channel.root_instrument._qumada_buffer for channel in self.gettable_channels if is_bufferable(channel)
            }
            self.trigger_ins = {
                param.root_instrument._qumada_mapping for param in self.dynamic_channels if is_triggerable(param)
            }
        self.sort_by_priority()
        self._lists_created = True
        self._relabel_instruments()

    def sort_by_priority(self):
        combined_lists = list(zip(self.dynamic_parameters, self.dynamic_channels, self.dynamic_sweeps))
        combined_sorted = sorted(combined_lists, key=lambda x: (x[0].get("priority", float("inf"))))
        if len(self.dynamic_parameters) > 0:
            self.dynamic_parameters, self.dynamic_channels, self.dynamic_sweeps = map(list, zip(*combined_sorted))
            for param in self.dynamic_parameters:
                try:
                    del param["priority"]
                except KeyError:
                    pass

    def initialize(self, dyn_ramp_to_val=False, inactive_dyn_channels: list | None = None) -> None:
        """
        Sets all static/sweepable parameters to their value/start value.
        If parameters are both, static and dynamic, they will be set to the "value" property
        and not to the "start" property.
        Parameters that are marked "dynamic" and "gettable" will not be added
        to the "self.gettable_parameters" as they are recorded anyway and will
        cause issues with dond functions.
        Provides gettable_parameters, static_parameters and dynamic parameters to
        measurement class and generates AbstractSweeps from the measurement
        properties. Sweeps form a list that can be found in "dynamic_sweeps"
        Relevant kwargs:
            dyn_ramp_to_val: Bool [False]: If true, dynamic parameters are
                    ramped to their value, before their sweep, else they are ramped
                    to their first setpoint.
            inactive_dyn_channels: List|None [None]: List of dynamic channels that are to be
                    treated as static for this initialization. They are always
                    ramped to their value instead of their sweeps starting point.
        """
        # TODO: Is there a more elegant way?
        # TODO: Put Sweep-Generation somewhere else?
        if inactive_dyn_channels is None:
            inactive_dyn_channels = []

        ramp_rate = self.settings.get("ramp_rate", 0.3)
        ramp_time = self.settings.get("ramp_time", 5)
        setpoint_intervall = self.settings.get("setpoint_intervall", 0.1)
        if not self._lists_created:
            self.generate_lists()
        # for item in self.compensated_parameters:
        #     if item not in self.dynamic_parameters:
        #         raise Exception(f"{item} is not in dynamic parameters and cannot be compensated!")
        # self.dynamic_sweeps = []
        self.compensating_sweeps = []
        for gate, parameters in self.gate_parameters.items():
            for parameter, channel in parameters.items():
                if self.properties[gate][parameter]["type"].find("static") >= 0:
                    ramp_or_set_parameter(
                        channel,
                        self.properties[gate][parameter]["value"],
                        ramp_rate=ramp_rate,
                        ramp_time=ramp_time,
                        setpoint_intervall=setpoint_intervall,
                    )
                elif self.properties[gate][parameter]["type"].find("dynamic") >= 0:
                    if self.properties[gate][parameter].get("_is_triggered", False) and self.buffered:
                        if "num_points" in self.properties[gate][parameter].keys():
                            try:
                                assert self.properties[gate][parameter]["num_points"] == self.buffered_num_points
                            except AssertionError:
                                logger.warning(
                                    f"Number of datapoints from buffer_settings\
                                    and gate_parameters do not match. Using \
                                    the value from the buffer settings: \
                                    {self.buffered_num_points}"
                                )
                                self.properties[gate][parameter]["num_points"] = self.buffered_num_points

                        elif "setpoints" in self.properties[gate][parameter].keys():
                            try:
                                assert len(self.properties[gate][parameter]["setpoints"]) == self.buffered_num_points
                            except AssertionError:
                                logger.warning(
                                    f"Number of datapoints from buffer_settings\
                                    and gate_parameters do not match. Using \
                                    the value from the buffer settings: \
                                    {self.buffered_num_points}"
                                )

                        else:
                            logger.info(
                                "No num_points or setpoints given for\
                                         buffered measurement. The value from \
                                         buffer_settings is used"
                            )
                    #     try:
                    #         self.dynamic_sweeps.append(
                    #             LinSweep(
                    #                 channel,
                    #                 self.properties[gate][parameter]["start"],
                    #                 self.properties[gate][parameter]["stop"],
                    #                 int(self.buffered_num_points),
                    #                 delay=self.properties[gate][parameter].setdefault("delay", 0),
                    #             )
                    #         )
                    #     except KeyError:
                    #         self.dynamic_sweeps.append(
                    #             CustomSweep(
                    #                 channel,
                    #                 self.properties[gate][parameter]["setpoints"],
                    #                 delay=self.properties[gate][parameter].setdefault("delay", 0),
                    #             )
                    #         )
                    # else:
                    #     try:
                    #         self.dynamic_sweeps.append(
                    #             LinSweep(
                    #                 channel,
                    #                 self.properties[gate][parameter]["start"],
                    #                 self.properties[gate][parameter]["stop"],
                    #                 int(self.properties[gate][parameter]["num_points"]),
                    #                 delay=self.properties[gate][parameter].setdefault("delay", 0),
                    #             )
                    #         )
                    #     except KeyError:
                    #         self.dynamic_sweeps.append(
                    #             CustomSweep(
                    #                 channel,
                    #                 self.properties[gate][parameter]["setpoints"],
                    #                 delay=self.properties[gate][parameter].setdefault("delay", 0),
                    #             )
                    #         )

                    # Handle different possibilities for starting points
                    if dyn_ramp_to_val or channel in inactive_dyn_channels:
                        try:
                            ramp_or_set_parameter(
                                channel,
                                self.properties[gate][parameter]["value"],
                                ramp_rate=ramp_rate,
                                ramp_time=ramp_time,
                                setpoint_intervall=setpoint_intervall,
                            )
                        except KeyError:
                            try:
                                ramp_or_set_parameter(
                                    channel,
                                    self.properties[gate][parameter]["start"],
                                    ramp_rate=ramp_rate,
                                    ramp_time=ramp_time,
                                    setpoint_intervall=setpoint_intervall,
                                )
                            except KeyError:
                                ramp_or_set_parameter(
                                    channel,
                                    self.properties[gate][parameter]["setpoints"][0],
                                    ramp_rate=ramp_rate,
                                    ramp_time=ramp_time,
                                    setpoint_intervall=setpoint_intervall,
                                )
                    else:
                        try:
                            ramp_or_set_parameter(
                                channel,
                                self.properties[gate][parameter]["start"],
                                ramp_rate=ramp_rate,
                                ramp_time=ramp_time,
                                setpoint_intervall=setpoint_intervall,
                            )
                        except KeyError:
                            ramp_or_set_parameter(
                                channel,
                                self.properties[gate][parameter]["setpoints"][0],
                                ramp_rate=ramp_rate,
                                ramp_time=ramp_time,
                                setpoint_intervall=setpoint_intervall,
                            )

                    # Generate sweeps from parameters
        self.active_compensated_channels = []
        self.active_compensating_channels = []
        self.active_compensating_parameters = []
        inactive_dyn_params = []
        for ch in inactive_dyn_channels:
            inactive_dyn_params.append(self.dynamic_parameters[self.dynamic_channels.index(ch)])
        for gate, parameters in self.gate_parameters.items():
            for parameter, channel in parameters.items():
                # This iterates over all compensating parameters
                if self.properties[gate][parameter]["type"].find("comp") >= 0:
                    try:
                        for i in range(len(self.compensating_parameters)):
                            if (
                                self.compensating_parameters[i]["gate"] == gate
                                and self.compensating_parameters[i]["parameter"] == parameter
                            ):
                                break
                        i = self.compensating_parameters.index({"gate": gate, "parameter": parameter})
                        leverarms = self.compensating_leverarms[i]
                        comped_params = copy.deepcopy(
                            self.compensated_parameters[i]
                        )  # list of parameters compensated by the current parameter
                        comped_sweeps = []  # Sweeps that are compensated by current param
                        comped_leverarms = []  # Leverarms of the current param
                        comping_sweeps = []  # List to store only the sweeps for the current param
                        k = 0
                        for comped_param in comped_params.copy():
                            # Check if the parameter is actually ramped in this part of the measurement
                            if comped_param in inactive_dyn_params:
                                comped_params.remove(comped_param)
                            else:
                                # Get only the relevant list entries for the current parameter
                                try:
                                    for i in range(len(self.dynamic_parameters)):
                                        if (
                                            self.dynamic_parameters[i]["gate"] == comped_param["gate"]
                                            and self.dynamic_parameters[i]["parameter"] == comped_param["parameter"]
                                        ):
                                            comped_index = i
                                            break
                                    # comped_index = self.dynamic_parameters.index(comped_param)
                                except ValueError as e:
                                    logger.exception(
                                        "Watch out, there is an Exception incoming!"
                                        + "Did you try to compensate for a not dynamic parameter?"
                                    )
                                    raise e
                                comped_sweeps.append(self.dynamic_sweeps[comped_index])
                                comped_leverarms.append(leverarms[k])
                                self.active_compensated_channels.append(self.dynamic_channels[comped_index])
                            k += 1
                        compensating_param = self.compensating_parameters[i]
                        self.active_compensating_parameters.append(compensating_param)
                        if len(comped_params) > 0:
                            self.active_compensating_channels.append(channel)
                            for j in range(len(comped_params)):
                                # Here we create lists/sweeps only containing the difference required for compensation.
                                # Still has to be substracted from the set value in the measurement script as this can
                                # depend on the measurement script used (e.g. 1D vs 2D sweeps)
                                comping_setpoints = (
                                    -1
                                    * float(comped_leverarms[j])
                                    * (np.array(comped_sweeps[j].get_setpoints()) - comped_sweeps[j].get_setpoints()[0])
                                )
                                # This creates an inner list of required setpoint differences only
                                # for the param that is currently iterated over!
                                # The final self.compensating_sweeps list will contain list for each
                                # compensating parameters with one sweep per
                                # parameter that is compensated by this compensating parameters.
                                comping_sweeps.append(
                                    CustomSweep(
                                        channel,
                                        comping_setpoints,
                                        delay=self.properties[gate][parameter].setdefault("delay", 0),
                                    )
                                )
                            self.compensating_sweeps.append(comping_sweeps)
                            if (
                                any(
                                    [
                                        self.properties[param["gate"]][param["parameter"]].get("_is_triggered", False)
                                        for param in comped_params
                                    ]
                                )
                                and self.buffered
                            ):
                                self.properties[compensating_param["gate"]][compensating_param["parameter"]][
                                    "_is_triggered"
                                ] = True
                            # TODO: This part has to be moved into the measurement script,
                            # as the final setpoints for the comping params are now set at
                            # the measurement script. A helper method would be nice to have.
                            # if min(self.compensating_sweeps[-1].get_setpoints()) < min(*self.compensating_limits[i]) \
                            #  or max(self.compensating_sweeps[-1].get_setpoints()) > max(*self.compensating_limits[i]):
                            #     raise Exception(f"Value for compensating gate {compensating_param} exceeds limits!")
                        ramp_or_set_parameter(
                            channel,
                            self.properties[gate][parameter]["value"],
                            ramp_rate=ramp_rate,
                            ramp_time=ramp_time,
                            setpoint_intervall=setpoint_intervall,
                        )
                    except ValueError as e:
                        raise e

        if self.buffered:
            for gettable_param in list(set(self.gettable_channels) - set(self.static_gettable_channels)):
                if is_bufferable(gettable_param):
                    gettable_param.root_instrument._qumada_buffer.subscribe([gettable_param])
                else:
                    raise Exception(f"{gettable_param} is not bufferable.")

    @abstractmethod
    def run(self) -> list:
        """
        Runs the already setup measurement. you can call self.initialize in here.
        Abstract method.
        """
        return []

    def reset(self) -> None:
        """
        Resets all static/dynamic parameters to their value/start value.
        TODO: Remove! Since initialize() does only create lists one, there is no advantage of using reset().
        """
        logger.warning(
            "The reset() method is deprecated and will be removed in a future release! \
                        It is recommended to replace all calls of reset() with initialize()"
        )
        ramp_rate = self.settings.get("ramp_rate", 0.3)
        setpoint_intervall = self.settings.get("setpoint_intervall", 0.1)
        for gate, parameters in self.gate_parameters.items():
            for parameter, channel in parameters.items():
                if self.properties[gate][parameter]["type"].find("static") >= 0:
                    ramp_or_set_parameter(
                        channel,
                        self.properties[gate][parameter]["value"],
                        ramp_rate=ramp_rate,
                        setpoint_intervall=setpoint_intervall,
                    )
                elif self.properties[gate][parameter]["type"].find("dynamic") >= 0:
                    try:
                        ramp_or_set_parameter(
                            channel,
                            self.properties[gate][parameter]["value"],
                            ramp_rate=ramp_rate,
                            setpoint_intervall=setpoint_intervall,
                        )
                    except KeyError:
                        try:
                            ramp_or_set_parameter(
                                channel,
                                self.properties[gate][parameter]["start"],
                                ramp_rate=ramp_rate,
                                setpoint_intervall=setpoint_intervall,
                            )
                        except KeyError:
                            ramp_or_set_parameter(
                                channel,
                                self.properties[gate][parameter]["setpoints"][0],
                                ramp_rate=ramp_rate,
                                setpoint_intervall=setpoint_intervall,
                            )

    def clean_up(self, additional_actions: list[Callable] | None = None, **kwargs) -> None:
        """
        Things to do after the measurement is complete. Cleans up subscribed paramteres for
        buffered measurements by default.
        TODO: Hook into measurement.run()

        Args:
            additional_actions (list[Callable], optional):
                List of functions to be called after the measurement is
                complete. Defaults to None.
        """
        for buffer in self.buffers:
            buffer.unsubscribe(buffer._subscribed_parameters)
        self.measurement_name = None
        if additional_actions:
            for action in additional_actions:
                action()

    def ready_buffers(self, **kwargs) -> None:
        """
        Setup all buffers registered in the measurement and start them.
        """
        for buffer in self.buffers:
            buffer.setup_buffer(settings=self.buffer_settings)
            buffer.start()
        for trigger in self.trigger_ins:
            trigger.setup_trigger_in(trigger_settings=self.buffer_settings)

    def readout_buffers(self, **kwargs) -> dict:
        """
        Readout all buffer and return the results as list of tuples
        (parameters, values) as required by qcodes measurement context manager.

        Args:
            **kwargs (dict):
                timestamps (bool): Set True if timestamp data is to be included
                    in the results. Not implemented yet.

        Returns:
            dict: Results, list with one tuple for each subscribed parameter.
            Tuple contains (parameter, measurement_data).

        """
        # TODO: Handle multiple bursts etc.
        data = {}
        results = []
        for buffer in self.buffers:
            buffer.stop()
            data[buffer] = buffer.read()
            for param in buffer._subscribed_parameters:
                results.append((param, flatten_array(data[buffer][param.name])))
        if kwargs.get("timestamps", False):
            results.append(flatten_array(data[list(data.keys())[0]]["timestamps"]))
        return results

    def _relabel_instruments(self) -> None:
        """
        Changes the labels of all instrument channels to the
        corresponding name defined in the measurement script.
        Has to be done after mapping!
        """
        for gate, parameters in self.gate_parameters.items():
            for key, parameter in parameters.items():
                parameter.label = f"{gate} {key}"

    def _add_current_datetime_to_metadata(self, *args, add_datetime_to_metadata: bool = True, **kwargs):
        if add_datetime_to_metadata:
            try:
                metadata = self.metadata
                metadata.add_datetime_to_metadata(datetime.now())
            except Exception as ex:
                print(f"Datetime could not be added to metadata: {ex}")

    def _add_data_to_metadata(self, *args, add_data_to_metadata: bool = True, **kwargs):
        # Add script and parameters to metadata
        if add_data_to_metadata:
            try:
                metadata = self.metadata
                cls = type(self)
                db_location = qc.config.core.db_location
                metadata.add_data_to_metadata(db_location, "sqlite3", f"{cls.__name__}Data")
            except Exception as ex:
                print(f"Data could not be added to metadata: {ex}")

    def _insert_metadata_into_db(self, *args, insert_metadata_into_db: bool = True, **kwargs):
        if insert_metadata_into_db:
            try:
                metadata = self.metadata
                metadata.save()
            except Exception as ex:
                print(f"Metadata could not inserted into database: {ex}")


class VirtualGate:
    """Virtual Gate"""

    def __init__(self):
        self._functions = []

    @property
    def functions(self):
        """List of equipment Functions, the virtual gate shall have."""
        return self._functions

    @functions.setter
    def functions(self, functions: MutableSequence):
        self._functions = functions


class CustomSweep(AbstractSweep):
    """
    Custom sweep from array of setpoints.

    Args:
        param: Qcodes parameter to sweep.
        setpoints: Array of setpoints.
        delay: Time in seconds between two consequtive sweep points
    """

    def __init__(
        self,
        param: ParameterBase,
        setpoints: np.ndarray,
        delay: float = 0,
        post_actions: ActionsT = (),
    ):
        self._param = param
        self._setpoints = setpoints
        self._num_points = len(setpoints)
        self._delay = delay
        self._post_actions = post_actions

    def get_setpoints(self) -> np.ndarray:
        """
        1D array of setpoints
        """
        return self._setpoints

    @property
    def param(self) -> ParameterBase:
        return self._param

    @property
    def delay(self) -> float:
        return self._delay

    @property
    def num_points(self) -> int:
        return self._num_points

    @property
    def post_actions(self) -> ActionsT:
        return self._post_actions



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe


from __future__ import annotations

import importlib.resources as res
from pathlib import Path


def import_resources(package: res.Package, ext: str = "*", recursive: bool = True) -> list[res.Resource]:
    """
    Loads resource files from a given package and returns them as list of Path-objects.

    Args:
        package: Package to load the resources from.
        ext: File extension, eg. "*.json". Defaults to "*".
        recursive: Shall resources also load from subdirectories? Defaults to True.

    Returns:
        List of the resources' paths.
    """
    path: Path = res.files(package)
    if recursive:
        return list(path.rglob(ext))
    else:
        return list(path.glob(ext))



# Copyright (c) 2023 JARA Institute for Quantum Information
#
# This file is part of QuMADA.
#
# QuMADA is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# QuMADA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# QuMADA. If not, see <https://www.gnu.org/licenses/>.
#
# Contributors:
# - Daniel Grothe
# - Till Huckemann

from __future__ import annotations

import numpy as np
from jsonschema import validate
from pyvisa import VisaIOError
from qcodes.instrument_drivers.stanford_research.SR830 import SR830
from qcodes.parameters import Parameter

from qumada.instrument.buffers.buffer import Buffer, BufferException


class SR830Buffer(Buffer):
    """Buffer for Stanford SR830"""

    ch1_names = ["X", "R", "X Noise", "aux_in1", "aux_in2"]
    ch2_names = ["Y", "Phase", "Y Noise", "aux_in3", "aux_in4"]

    AVAILABLE_TRIGGERS: list[str] = ["external", "trig_in_1"]

    def __init__(self, device: SR830):
        self._device = device
        self._trigger: str | None = None
        self._subscribed_parameters: set[Parameter] = set()
        self._num_points: int | None = None

    def setup_buffer(self, settings: dict) -> None:
        """Sets instrument related settings for the buffer."""
        # TODO: Validation for sampling rates (look up in manual) Comment: Is this required? Is handled by driver...
        # TODO: Trigger und SR abgleichen
        # TODO: Are there different trigger modes?

        validate(settings, self.settings_schema)
        self.settings: dict = settings
        self._device.buffer_SR(settings.get("sampling_rate", 512))
        self._device.buffer_trig_mode("OFF")
        self._set_num_points()
        self.delay_data_points = 0  # Datapoints to delete at the beginning of dataset due to delay.
        self.delay = settings.get("delay", 0)
        if self.delay < 0:
            raise BufferException("The SR830'S Trigger Input does not support negative delays.")
        else:
            self.delay_data_points = int(self.delay * self._device.buffer_SR())
            self.num_points = self.delay_data_points + self.num_points
            # TODO: There has to be a more elegant way for the setter.

    @property
    def num_points(self) -> int | None:
        return self._num_points

    @num_points.setter
    def num_points(self, num_points) -> None:
        if num_points > 16383:
            raise BufferException(
                "SR830 is to small for this measurement. Please reduce the number of data points or the delay"
            )
        self._num_points = int(num_points)

    def _set_num_points(self) -> None:
        """
        Calculates number of datapoints and sets
        the num_points accordingly.
        Raises
        ------
        Exception
           Exception if number of points is overdefined.

        Returns
        -------
        None
        """
        if all(k in self.settings for k in ("sampling_rate", "burst_duration", "num_points")):
            raise BufferException("You cannot define sampling_rate, burst_duration and num_points at the same time")
        elif self.settings.get("num_points", False):
            self.num_points = self.settings["num_points"]
            if self.settings.get("sampling_rate", False):
                self._device.buffer_SR(self.settings["sampling_rate"])
            else:
                self._device.buffer_SR(self.num_points / self.settings["burst_duration"])
        elif all(k in self.settings for k in ("sampling_rate", "burst_duration")):
            self.num_points = int(np.ceil(self.settings["sampling_rate"] * self.settings["burst_duration"]))

    @property
    def trigger(self) -> str | None:
        return self._trigger

    @trigger.setter
    def trigger(self, trigger: str | None) -> None:
        if trigger == "trig_in_1":
            # TODO: standard value for Sample Rate
            self._device.buffer_SR(512)
            self._device.buffer_trig_mode("OFF")
        elif trigger == "external":
            self._device.buffer_SR("Trigger")
            self._device.buffer_trig_mode("ON")
        else:
            raise BufferException(
                "SR830 does not support setting custom trigger inputs. "
                "Use 'external' and the input on the back of the unit."
            )
        self._trigger = trigger

    def force_trigger(self) -> None:
        self._device.buffer_start()

    def read_raw(self) -> dict:
        # TODO: Handle stopping buffer or not
        data = {}
        try:
            for parameter in self._subscribed_parameters:
                if parameter.name in self.ch1_names:
                    ch = "ch1"
                elif parameter.name in self.ch2_names:
                    ch = "ch2"

                # TODO: what structure has the data? do we get timestamps?
                data[parameter.name] = self._device.__getattr__(f"{ch}_datatrace").get()
                data[parameter.name] = data[parameter.name][self.delay_data_points : self.num_points]
        except VisaIOError as ex:
            raise BufferException("Could not read the buffer. Buffer has to be stopped before readout.") from ex
        return data

    def read(self) -> dict:
        # TODO: Add timetrace if possible
        return self.read_raw()

    def subscribe(self, parameters: set | list[Parameter]) -> None:
        for parameter in parameters:
            name = parameter.name
            if name in self.ch1_names:
                self._device.ch1_display(name)
                param_to_remove = {param for param in self._subscribed_parameters if param.name in self.ch1_names}
                self._subscribed_parameters.difference_update(
                    param_to_remove
                )  # remove previously subscribed parameter from ch1
                self._subscribed_parameters.add(parameter)
            elif name in self.ch2_names:
                self._device.ch2_display(name)
                param_to_remove = {param for param in self._subscribed_parameters if param.name in self.ch2_names}
                self._subscribed_parameters.difference_update(
                    param_to_remove
                )  # remove previously subscribed parameter from ch2
                self._subscribed_parameters.add(parameter)
            else:
                raise BufferException(f"Parameter {parameter.name} can not be buffered.")

    def unsubscribe(self, parameters: set | list[Parameter]) -> None:
        for parameter in parameters.copy():
            name = parameter.name
            if name in ["X", "R", "X Noise", "aux_in1", "aux_in2"]:
                self._subscribed_parameters.remove(parameter)
            elif name in ["Y", "Phase", "Y Noise", "aux_in3", "aux_in4"]:
                self._subscribed_parameters.remove(parameter)
            else:
                raise BufferException(f"Parameter {parameter.name} can not be buffered.")

    def is_subscribed(self, parameter: Parameter) -> bool:
        return parameter in self._subscribed_parameters

    def start(self) -> None:
        self._device.buffer_reset()
        self._device.buffer_start()

    def stop(self) -> None:
        self._device.buffer_pause()

    def is_ready(self) -> bool: ...

    def is_finished(self) -> bool:
        if self._device.buffer_npts() >= self.num_points:
            self.stop()
            return True
        else:
            return False
